from picoscope import ps5000a
import matplotlib.pyplot as plt
import numpy as np
import time

ps = ps5000a.PS5000a()

# rapid block mode

ps.setChannel(channel="A", coupling="DC", VRange=1)
ps.setChannel(channel="B", enabled=False)

n_captures = 100
sample_interval = 100e-9  # 100 ns
sample_duration = 2e-3  # 1 ms
ps.setResolution('16')
ps.setSamplingInterval(sample_interval, sample_duration)
ps.setSimpleTrigger("A", threshold_V=0.1, timeout_ms=1)

samples_per_segment = ps.memorySegments(n_captures)
ps.setNoOfCaptures(n_captures)

data = np.zeros((n_captures, samples_per_segment), dtype=np.int16)

t1 = time.time()

ps.runBlock()
ps.waitReady()

t2 = time.time()
print("Time to record data to scope: ", str(t2 - t1))

ps.getDataRawBulk(data=data)

t3 = time.time()
print("Time to copy to RAM: ", str(t3 - t2))

plt.imshow(data[:, 0:ps.noSamples], aspect='auto', interpolation='none',
           cmap=plt.cm.hot)
plt.colorbar()
plt.show()

ps.close()



"""
This is the low level driver file for a specific Picoscope.

By this, I mean if parameters want to get passed as strings, they should be
handled by PSBase
All functions here should take things as close to integers as possible, the
only exception here is for array parameters. Array parameters should be passed
in a pythonic way through numpy since the PSBase class should not be aware of
the specifics behind how the clib is called.

The functions should not have any default values as these should be handled
by PSBase.
"""

# This is the instrument-specific file for the PS2000a series of instruments.
#
# pico-python is Copyright (c) 2013-2016 By:
# Colin O'Flynn <coflynn@newae.com>
# Mark Harfouche <mark.harfouche@gmail.com>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import math
import numpy as np

# to load the proper dll
import platform

# Do not import or use ill definied data types
# such as short int or long
# use the values specified in the h file
# float is always defined as 32 bits
# double is defined as 64 bits
from ctypes import byref, POINTER, create_string_buffer, c_float, \
    c_int16, c_int32, c_uint16, c_uint32, c_void_p
from ctypes import c_int32 as c_enum

from picoscope.picobase import _PicoscopeBase


class PS2000a(_PicoscopeBase):
    """The following are low-level functions for the PS2000a."""

    LIBNAME = "ps2000a"

    NUM_CHANNELS = 4
    CHANNELS = {"A": 0, "B": 1, "C": 2, "D": 3,
                "External": 4, "MaxChannels": 4, "TriggerAux": 5}

    ADC_RESOLUTIONS = {"8": 0, "12": 1, "14": 2, "15": 3, "16": 4}

    CHANNEL_RANGE = [{"rangeV": 10E-3, "apivalue": 0, "rangeStr": "10 mV"},
                     {"rangeV": 20E-3, "apivalue": 1, "rangeStr": "20 mV"},
                     {"rangeV": 50E-3, "apivalue": 2, "rangeStr": "50 mV"},
                     {"rangeV": 100E-3, "apivalue": 3, "rangeStr": "100 mV"},
                     {"rangeV": 200E-3, "apivalue": 4, "rangeStr": "200 mV"},
                     {"rangeV": 500E-3, "apivalue": 5, "rangeStr": "500 mV"},
                     {"rangeV": 1.0, "apivalue": 6, "rangeStr": "1 V"},
                     {"rangeV": 2.0, "apivalue": 7, "rangeStr": "2 V"},
                     {"rangeV": 5.0, "apivalue": 8, "rangeStr": "5 V"},
                     {"rangeV": 10.0, "apivalue": 9, "rangeStr": "10 V"},
                     {"rangeV": 20.0, "apivalue": 10, "rangeStr": "20 V"},
                     {"rangeV": 50.0, "apivalue": 11, "rangeStr": "50 V"},
                     ]

    CHANNEL_COUPLINGS = {"DC": 1, "AC": 0}

    # has_sig_gen = True
    WAVE_TYPES = {"Sine": 0, "Square": 1, "Triangle": 2,
                  "RampUp": 3, "RampDown": 4,
                  "Sinc": 5, "Gaussian": 6, "HalfSine": 7, "DCVoltage": 8,
                  "WhiteNoise": 9}

    SWEEP_TYPES = {"Up": 0, "Down": 1, "UpDown": 2, "DownUp": 3}

    SIGGEN_TRIGGER_TYPES = {"Rising": 0, "Falling": 1,
                            "GateHigh": 2, "GateLow": 3}
    SIGGEN_TRIGGER_SOURCES = {"None": 0, "ScopeTrig": 1, "AuxIn": 2,
                              "ExtIn": 3, "SoftTrig": 4, "TriggerRaw": 5}

    # TIME_UNITS = {'PS2000A_FS':0,'PS2000A_PS':1,
    # 'PS2000A_NS':2,'PS2000A_US':3,'PS2000A_MS':4,
    # 'PS2000A_S':5,'PS2000A_MAX_TIME_UNITS':6}
    TIME_UNITS = {0: 1e-15, 1: 1e-12, 2: 1e-9, 3: 1e-6, 4: 1e-3, 5: 1e0}

    AWGPhaseAccumulatorSize = 32

    # AWG scaling according to programming manual p.72
    # Vout = 1uV * (pkToPk/2) * (sample_value / 32767) + offsetVoltage
    # The API datatype is a (signed) short
    AWGMaxVal = 32767
    AWGMinVal = -32767

    AWG_INDEX_MODES = {"Single": 0, "Dual": 1, "Quad": 2}

    MAX_VALUE_8BIT = 32512
    MIN_VALUE_8BIT = -32512
    MAX_VALUE_OTHER = 32767
    MIN_VALUE_OTHER = -32767

    EXT_RANGE_VOLTS = 5

    def __init__(self, serialNumber=None, connect=True):
        """Load DLL etc."""
        if platform.system() == 'Linux':
            from ctypes import cdll
            self.lib = cdll.LoadLibrary("lib" + self.LIBNAME + ".so")
        elif platform.system() == 'Darwin':
            self.lib = self.loadLibraryDarwin("lib" + self.LIBNAME + ".dylib")
        else:
            from ctypes import windll
            from ctypes.util import find_library
            self.lib = windll.LoadLibrary(
                find_library(str(self.LIBNAME + ".dll"))
            )

        self.resolution = self.ADC_RESOLUTIONS["8"]

        super(PS2000a, self).__init__(serialNumber, connect)

    def _lowLevelOpenUnit(self, sn):
        c_handle = c_int16()
        if sn is not None:
            serialNullTermStr = byref(create_string_buffer(sn))
        else:
            serialNullTermStr = None
        # Passing None is the same as passing NULL
        m = self.lib.ps2000aOpenUnit(byref(c_handle), serialNullTermStr)
        self.checkResult(m)
        self.handle = c_handle.value

        # The scaling factor used in the timebase calculation varies based on
        # the particular model. See section 2.8 (pg 27) of the 2000a
        # programmer's guide
        self.model = self.getUnitInfo('VariantInfo')
        if self.model in ('2205AMSO', '2206', '2206A', '2206B', '2405A'):
            # 500 MS/s
            self._timebase_to_timestep = \
                lambda n: (2**n / 5e8) if n < 3 else ((n - 2) / 625e5)
            self._timestep_to_timebase = \
                lambda t: math.log(t * 5e8, 2) if t < 16e-9 else (
                    (t * 625e5) + 2)
        elif self.model in ('2206BMSO', '2207', '2207A', '2207B', '2207BMSO',
                            '2208', '2208A', '2208B', '2208BMSO', '2406B',
                            '2407B', '2408B'):
            # 1 GS/s
            self._timebase_to_timestep = lambda n: (2**n / 1e9) if n < 3 else (
                (n - 2) / 125e6)
            self._timestep_to_timebase = \
                lambda t: math.log(t * 1e9, 2) if t < 8e-9 else (
                    (t * 125e6) + 2)
        elif self.model == '2205MSO':
            self._timebase_to_timestep = \
                lambda n: (2**n / 2e8) if n < 1 else (n / 1e8)
            self._timestep_to_timebase = \
                lambda t: math.log(t * 2e8, 2) if t < 10e-9 else (t * 1e8)
        else:
            raise ValueError("Unrecognised variant {}".format(self.model))

        # The AWG parameters vary based on the particular model. See section
        # 3.51.2 of the 2000a programmer's guide
        if self.model in ('2205AMSO', '2206', '2206A', '2207', '2207A', '2208',
                          '2208A', '2405A'):
            self.AWGBufferAddressWidth = 13
            self.AWGDACInterval = 50E-9
        elif self.model in ('2206B', '2206BMSO', '2207B', '2207BMSO', '2208B',
                            '2208BMSO', '2406B', '2407B', '2408B'):
            self.AWGBufferAddressWidth = 15
            self.AWGDACInterval = 50E-9
        else:
            # The programmer's manual indicates that some older models have
            # these parameters. Just use them as a catch-all for any models
            # not listed above
            self.AWGBufferAddressWidth = 13
            self.AWGDACInterval = 20.83E-9

        self.AWGMaxSamples = 2 ** self.AWGBufferAddressWidth
        self.AWGDACFrequency = 1 / self.AWGDACInterval

    def _lowLevelCloseUnit(self):
        m = self.lib.ps2000aCloseUnit(c_int16(self.handle))
        self.checkResult(m)

    def _lowLevelSetChannel(self, chNum, enabled, coupling, VRange, VOffset,
                            BWLimited):
        m = self.lib.ps2000aSetChannel(c_int16(self.handle), c_enum(chNum),
                                       c_int16(enabled), c_enum(coupling),
                                       c_enum(VRange), c_float(VOffset))
        self.checkResult(m)

    def _lowLevelStop(self):
        m = self.lib.ps2000aStop(c_int16(self.handle))
        self.checkResult(m)

    def _lowLevelGetUnitInfo(self, info):
        s = create_string_buffer(256)
        requiredSize = c_int16(0)

        m = self.lib.ps2000aGetUnitInfo(c_int16(self.handle), byref(s),
                                        c_int16(len(s)), byref(requiredSize),
                                        c_enum(info))
        self.checkResult(m)
        if requiredSize.value > len(s):
            s = create_string_buffer(requiredSize.value + 1)
            m = self.lib.ps2000aGetUnitInfo(c_int16(self.handle), byref(s),
                                            c_int16(len(s)),
                                            byref(requiredSize), c_enum(info))
            self.checkResult(m)

        # should this bee ascii instead?
        # I think they are equivalent...
        return s.value.decode('utf-8')

    def _lowLevelFlashLed(self, times):
        m = self.lib.ps2000aFlashLed(c_int16(self.handle), c_int16(times))
        self.checkResult(m)

    def _lowLevelSetSimpleTrigger(self, enabled, trigsrc, threshold_adc,
                                  direction, delay, auto):
        m = self.lib.ps2000aSetSimpleTrigger(
            c_int16(self.handle), c_int16(enabled),
            c_enum(trigsrc), c_int16(threshold_adc),
            c_enum(direction), c_uint32(delay), c_int16(auto))
        self.checkResult(m)

    def _lowLevelSetNoOfCaptures(self, numCaptures):
        m = self.lib.ps2000aSetNoOfCaptures(
            c_int16(self.handle), c_uint16(numCaptures))
        self.checkResult(m)

    def _lowLevelMemorySegments(self, numSegments):
        maxSamples = c_int32()
        m = self.lib.ps2000aMemorySegments(c_int16(self.handle),
                                           c_uint16(numSegments),
                                           byref(maxSamples))
        self.checkResult(m)
        return maxSamples.value

    def _lowLevelGetMaxSegments(self):
        maxSegments = c_int16()
        m = self.lib.ps2000aGetMaxSegments(c_int16(self.handle),
                                           byref(maxSegments))
        self.checkResult(m)
        return maxSegments.value

    def _lowLevelRunBlock(self, numPreTrigSamples, numPostTrigSamples,
                          timebase, oversample, segmentIndex, callback,
                          pParameter):
        # NOT: Oversample is NOT used!
        timeIndisposedMs = c_int32()
        m = self.lib.ps2000aRunBlock(
            c_int16(self.handle), c_uint32(numPreTrigSamples),
            c_uint32(numPostTrigSamples), c_uint32(timebase),
            c_int16(oversample), byref(timeIndisposedMs),
            c_uint32(segmentIndex),
            c_void_p(), c_void_p())
        # According to the documentation, 'callback, pParameter' should work
        # instead of the last two c_void_p parameters.
        self.checkResult(m)
        return timeIndisposedMs.value

    def _lowLevelIsReady(self):
        ready = c_int16()
        m = self.lib.ps2000aIsReady(c_int16(self.handle), byref(ready))
        self.checkResult(m)
        if ready.value:
            return True
        else:
            return False

    def _lowLevelGetTimebase(self, tb, noSamples, oversample, segmentIndex):
        """Return (timeIntervalSeconds, maxSamples)."""
        maxSamples = c_int32()
        intervalNanoSec = c_float()

        m = self.lib.ps2000aGetTimebase2(
            c_int16(self.handle), c_uint32(tb), c_uint32(noSamples),
            byref(intervalNanoSec), c_int16(oversample), byref(maxSamples),
            c_uint32(segmentIndex))
        self.checkResult(m)
        # divide by 1e9 to return interval in seconds
        return (intervalNanoSec.value * 1e-9, maxSamples.value)

    def getTimeBaseNum(self, sampleTimeS):
        """Convert sample time in S to something to pass to API Call."""
        clipped = np.clip(math.floor(self._timestep_to_timebase(sampleTimeS)),
                          0, np.iinfo(np.int32).max)

        return int(clipped)

    def getTimestepFromTimebase(self, timebase):
        """Convvert API timestep code to sampling interval.

        API timestep as an integer from 0-32,
        sampling interval in seconds.
        """
        return self._timebase_to_timestep(timebase)

    def _lowLevelSetAWGSimpleDeltaPhase(self, waveform, deltaPhase,
                                        offsetVoltage, pkToPk, indexMode,
                                        shots, triggerType, triggerSource):
        """Waveform should be an array of shorts."""
        waveformPtr = waveform.ctypes.data_as(POINTER(c_int16))

        m = self.lib.ps2000aSetSigGenArbitrary(
            c_int16(self.handle),
            c_uint32(int(offsetVoltage * 1E6)),  # offset voltage in microvolts
            c_uint32(int(pkToPk * 1E6)),         # pkToPk in microvolts
            c_uint32(int(deltaPhase)),           # startDeltaPhase
            c_uint32(int(deltaPhase)),           # stopDeltaPhase
            c_uint32(0),                         # deltaPhaseIncrement
            c_uint32(0),                         # dwellCount
            waveformPtr,                         # arbitraryWaveform
            c_int32(len(waveform)),              # arbitraryWaveformSize
            c_enum(0),                           # sweepType for deltaPhase
            c_enum(0),            # operation (adding random noise and whatnot)
            c_enum(indexMode),                   # single, dual, quad
            c_uint32(shots),
            c_uint32(0),                         # sweeps
            c_uint32(triggerType),
            c_uint32(triggerSource),
            c_int16(0))                          # extInThreshold
        self.checkResult(m)

    def _lowLevelSetDataBuffer(self, channel, data, downSampleMode,
                               segmentIndex):
        """Set the buffer for the picoscope.

        Be sure to call _lowLevelClearDataBuffer
        when you are done with the data array
        or else subsequent calls to GetValue will still use the same array.
        """
        dataPtr = data.ctypes.data_as(POINTER(c_int16))
        numSamples = len(data)

        m = self.lib.ps2000aSetDataBuffer(
            c_int16(self.handle), c_enum(channel), dataPtr,
            c_int32(numSamples), c_uint32(segmentIndex),
            c_enum(downSampleMode))
        self.checkResult(m)

    def _lowLevelSetMultipleDataBuffers(self, channel, data, downSampleMode):
        max_segments = self._lowLevelGetMaxSegments()
        if data.shape[0] < max_segments:
            raise ValueError(
                "data array has fewer rows" +
                " than current number of memory segments")
        if data.shape[1] < self.maxSamples:
            raise ValueError("data array has fewer columns than maxSamples")

        for i in range(max_segments):
            m = self._lowLevelSetDataBuffer(channel, data[i, :],
                                            downSampleMode, i)
            self.checkResult(m)

    def _lowLevelClearDataBuffer(self, channel, segmentIndex):
        """Clear the data in the picoscope."""
        m = self.lib.ps2000aSetDataBuffer(
            c_int16(self.handle), c_enum(channel),
            c_void_p(), c_uint32(0), c_uint32(segmentIndex), c_enum(0))
        self.checkResult(m)

    def _lowLevelGetValues(self, numSamples, startIndex, downSampleRatio,
                           downSampleMode, segmentIndex):
        numSamplesReturned = c_uint32()
        numSamplesReturned.value = numSamples
        overflow = c_int16()
        m = self.lib.ps2000aGetValues(
            c_int16(self.handle), c_uint32(startIndex),
            byref(numSamplesReturned), c_uint32(downSampleRatio),
            c_enum(downSampleMode), c_uint32(segmentIndex),
            byref(overflow))
        self.checkResult(m)
        return (numSamplesReturned.value, overflow.value)

    def _lowLevelGetValuesBulk(self, numSamples, fromSegment, toSegment,
                               downSampleRatio, downSampleMode, overflow):
        m = self.lib.ps2000aGetValuesBulk(
            c_int16(self.handle),
            byref(c_int16(numSamples)),
            c_int16(fromSegment),
            c_int16(toSegment),
            c_int32(downSampleRatio),
            c_int16(downSampleMode),
            overflow.ctypes.data_as(POINTER(c_int16))
            )
        self.checkResult(m)
        return overflow, numSamples

    def _lowLevelGetTriggerTimeOffset(self, segmentIndex):
        timeUpper = c_uint32()
        timeLower = c_uint32()
        timeUnits = c_int16()
        m = self.lib.ps2000aGetTriggerTimeOffset(
            c_int16(self.handle),
            byref(timeUpper),
            byref(timeLower),
            byref(timeUnits),
            c_uint32(segmentIndex),
            )
        self.checkResult(m)

        # timeUpper and timeLower are the upper 4 and lower 4 bytes of a 64-bit
        # (8-byte) integer which is scaled by timeUnits to get the precise
        # trigger location
        return (((timeUpper.value << 32) + timeLower.value) *
                self.TIME_UNITS[timeUnits.value])

    def _lowLevelSetSigGenBuiltInSimple(self, offsetVoltage, pkToPk, waveType,
                                        frequency, shots, triggerType,
                                        triggerSource, stopFreq, increment,
                                        dwellTime, sweepType, numSweeps):
        # TODO, I just noticed that V2 exists
        # Maybe change to V2 in the future

        if stopFreq is None:
            stopFreq = frequency

        m = self.lib.ps2000aSetSigGenBuiltIn(
            c_int16(self.handle),
            c_int32(int(offsetVoltage * 1000000)),
            c_int32(int(pkToPk * 1000000)),
            c_int16(waveType),
            c_float(frequency), c_float(stopFreq),
            c_float(increment), c_float(dwellTime),
            c_enum(sweepType), c_enum(0),
            c_uint32(shots), c_uint32(numSweeps),
            c_enum(triggerType), c_enum(triggerSource),
            c_int16(0))
        self.checkResult(m)

    def _lowLevelSigGenSoftwareControl(self, triggerType):
        m = self.lib.ps2000aSigGenSoftwareControl(
            c_int16(self.handle), c_enum(triggerType))
        self.checkResult(m)



import matplotlib.pyplot as plt
import numpy as np
import time

from picoscope import ps3000a

SERIAL_NUM = 'AR911/011\x00'
ps = ps3000a.PS3000a(SERIAL_NUM)

# now = time.strftime("%Y%m%d_%H%M%S")
# filename = "sweep_" + now + ".swp"
# output_file = open(filename, "wb")

c = 3e8

# rapid block mode

ps.setChannel(channel="A", coupling="DC", VRange=1)

n_captures = 2300 * 3  # int(600 * 1.4)
sample_interval = 5 / 3e8
sample_duration = 1e3 * 2 / 3e8

ps.setSamplingInterval(sample_interval, sample_duration)
ps.setSimpleTrigger("A", threshold_V=0.2)

samples_per_segment = ps.memorySegments(n_captures)
ps.setNoOfCaptures(n_captures)

data = np.zeros((n_captures, samples_per_segment), dtype=np.int16)

t1 = time.time()

ps.runBlock()
ps.waitReady()

t2 = time.time()
print("Time to get sweep: " + str(t2 - t1))

ps.getDataRawBulk(data=data)

# for i in range(n_captures):
#     ps._lowLevelSetDataBuffer(ps.CHANNELS["A"],
#         data[i, :], 0, i)

# # t2 = time.time()
# nsamples = c_int32(ps.noSamples)
# from_segment_index = 0
# to_segment_index = n_captures - 1
# downsample_ratio = 0
# downsample_mode = 0
# overflow = np.zeros(n_captures, dtype=np.int16)
# overflow_ptr = overflow.ctypes.data_as(POINTER(c_int16))

# m = ps.lib.ps3000aGetValuesBulk(c_int16(ps.handle),
#         byref(nsamples),
#         c_int16(from_segment_index),
#         c_int16(to_segment_index),
#         c_int32(downsample_ratio),
#         c_int16(downsample_mode),
#         overflow_ptr)
# print m

# ps.checkResult(m)

t3 = time.time()
print("Time to read data: " + str(t3 - t2))


# output_file.write(data)
# t4 = time.time()
# print "Time to write data to disk: ", str(t4 - t3)
# output_file.close()

plt.imshow(data[:, 0:ps.noSamples], aspect='auto', interpolation='none',
           cmap=plt.cm.hot)
plt.colorbar()
plt.show()

ps.close()



from ctypes import c_short, byref, cdll, c_char, create_string_buffer
import time
import platform

if platform.system() == "Windows":
    from ctypes import windll

    lib = windll.LoadLibrary("ps3000a.dll")
elif platform.system() == "Linux":
    lib = cdll.LoadLibrary("libps3000a.so")


# Enumerating units
print("Enumerating units...\n\n")

count = c_short()
serials = (c_char * 100)()
serialLth = c_short()
lib.ps3000aEnumerateUnits(byref(count), byref(serials), byref(serialLth))

print("count:" + str(count))
print("serial numbers: " + str(serials[0:100]))
print("serial string length: " + str(serialLth))

# try connecting
serial_len = serialLth.value
serial_num = serials[0:serial_len]
print("\n\nAttempting to connect to" + str(serial_num) + "\n\n")
handle = c_short()
s = create_string_buffer(serial_num)

m = lib.ps3000aOpenUnit(byref(handle), byref(s))
print("Result code:" + str(m))
print("Scope handle:" + str(handle.value))

try:
    # flash led
    print("\n\nFlashing LED...\n\n")
    start = c_short(10)
    lib.ps3000aFlashLed(handle, start)
    time.sleep(5)
except:  # noqa
    print("led wouldn't flash")
    pass
# close unit
print("\n\nClosing scope...\n\n")
m = lib.ps3000aCloseUnit(handle)
if m == 0:
    print("Unit" + str(serial_num) + "shut down.")



# -*- coding: utf-8 -*-
#
# Based on code from Luke Campagnola, author of amazing pyqtgraph library

from PyQt4 import QtGui, QtCore
import pyqtgraph as pg
import numpy as np
from picoscope import ps6000


def setupScope():
    ps = ps6000.PS6000()

    # Example of simple capture
    res = ps.setSamplingFrequency(500E6, 4096)
    sampleRate = res[0]
    print("Sampling @ %f MHz, %d samples" % (res[0] / 1E6, res[1]))
    ps.setChannel("A", "AC", 50E-3)
    return [ps, sampleRate]


class MainWindow(QtGui.QMainWindow):
    def __init__(self):
        QtGui.QMainWindow.__init__(self)

        self.plotLayout = pg.GraphicsLayoutWidget()
        self.setCentralWidget(self.plotLayout)

        self.plot = self.plotLayout.addPlot(row=0, col=0)
        # self.avgPlot = self.plotLayout.addPlot(row=1, col=0)
        self.specPlot = self.plotLayout.addPlot(
            row=1, col=0, labels={'bottom': ('Frequency', 'Hz')})
        self.specPlot.setYRange(0, 0.2)

        self.wave = ScrollingPlot()
        self.plot.addItem(self.wave)

        self.triggerY = pg.InfiniteLine(angle=0, movable=True)
        self.triggerX = pg.InfiniteLine(movable=True)
        self.plot.addItem(self.triggerX)
        self.plot.addItem(self.triggerY)

        self.plot.setXRange(-1000, 1000)
        self.plot.setYRange(-0.05, 0.05)

        self.resize(800, 800)
        self.show()

        [self.scope, self.rate] = setupScope()
        self.scopeRunning = False

        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self.update)
        self.timer.start(50)

        self.trigAvg = np.zeros(5000)
        self.lastValue = None
        self.lastData = None
        self.persistent = []
        self.lastUpdate = None

    def update(self):
        now = pg.ptime.time()
        if self.lastUpdate is None:
            self.lastUpdate = now
            dt = 0.0
        else:
            dt = now - self.lastUpdate
            self.lastUpdate = now

        # read data from sound device
        if self.scopeRunning is False:
            self.scope.runBlock()
            self.scopeRunning = True

        if self.scope.isReady():
            data = self.scope.getDataV(0, 4096)
            self.scopeRunning = False
        else:
            data = []

        if self.scopeRunning is False:
            self.scope.runBlock()

        # if self.lastData is not None:
        #     data.append([self.lastData[-1]])

        """
        while True:
            chunk = np.fromstring(self.pcm.read()[1], dtype=np.int16) / 2.0**15
            if len(chunk) == 0 and sum(map(len, data)) > 1024:
                break
            data.append(chunk)
        """

        if len(data) > 0:
            # data = np.concatenate(data)
            self.wave.append(data)

            # If there is a trigger available, shift wave plot to align
            # and add a persistent trace
            ty = self.triggerY.value()
            tx = self.triggerX.value()

            # if self.lastData is None:
            #     trigData = data
            # else:
            #     trigData = np.concatenate([self.lastData[:-15], data])
            tind = np.argwhere((data[:-15] < ty) * (data[15:] >= ty))
            if len(tind) > 0:
                tind = tind[min(2, tind.shape[0] - 1), 0] + 15
                self.wave.setPos(data.shape[0] - tind + tx, 0)

                # update persistent
                if len(self.wave.plots) > 1:
                    for i in [1, 2]:
                        d = self.wave.plots[-i].yData
                        p = pg.PlotDataItem(d)
                        self.persistent.append(p)
                        self.plot.addItem(p)
                        p.setPos(self.wave.plots[-i].pos() + self.wave.pos())

            else:
                self.wave.setPos(data.shape[0], 0)

            # update spectrum
            spec = np.abs(np.fft.fft(data[-1024:])[:512])
            x = np.linspace(0, self.rate / 2., len(spec))
            self.specPlot.plot(x, spec, clear=True)
            self.lastData = data

        while len(self.persistent) > 30:  # limit # of persistent plots
            p = self.persistent.pop(0)
            self.plot.removeItem(p)
        for p in self.persistent[:]:
            p.setOpacity(p.opacity() * (0.03**dt))
            if p.opacity() < 0.01:
                self.persistent.remove(p)
                self.plot.removeItem(p)

    def keyPressEvent(self, ev):
        if ev.text() == ' ':
            if self.timer.isActive():
                self.timer.stop()
            else:
                self.timer.start()
        else:
            print(ev.key())


class ScrollingPlot(QtGui.QGraphicsItem):
    """Simple appendable plot.

    New data is appended to the right and shifts existing data leftward."""
    def __init__(self):
        QtGui.QGraphicsItem.__init__(self)
        self.setFlag(self.ItemHasNoContents)

        self.plots = []
        self.lastValue = 0

    def append(self, data):
        # remove plots that are too old
        scene = self.scene()
        while len(self.plots) > 20:
            scene.removeItem(self.plots.pop(0))

        # add the next plot, shift to its correct position
        p = pg.PlotDataItem(data)
        self.plots.append(p)
        p.setParentItem(self)

        shift = len(data) - 1
        for p in self.plots:
            p.moveBy(-shift, 0)

    def boundingRect(self):
        return QtCore.QRectF()


if __name__ == '__main__':
    app = QtGui.QApplication([])
    win = MainWindow()
    app.exec_()



# -*- coding: utf-8 -*-
"""
Testing the ps6000a series software-device interaction.
======================================================

This file provides tests in order to verify that the software does with the
oscilloscope, what it is supposed to do. Additionally the tests serve as
examples on how to do certain tasks.

Usage
-----

- Run this file in order to execute all software tests on your device.
- Import this file and execute the tests you want with your already opened
  oscilloscope device.

Created on Tue Jan 25 12:20:14 2022 by Benedikt Moneke
"""

# Imports
import matplotlib.pyplot as plt
import numpy as np
from picoscope import ps6000a
import time


def test_general_unit_calls(ps):
    """Test general unit calls"""
    ps.flashLed()
    print(ps.getAllUnitInfo())
    assert not ps.ping(), "Ping failed."
    print("General unit calls test passed.")


def test_timebase(ps):
    """Test the timebase methods."""
    # Preparation
    ps.memorySegments(1)
    # Tests
    data = ((200e-12, 0),
            (800e-12, 2),
            (3.2e-9, 4),
            (6.4e-9, 5),
            (10e-9, 5),
            (15e-9, 6))
    for (t, timebase) in data:
        text = "Time {} does not give timebase {}".format(t, timebase)
        assert ps.getTimeBaseNum(t) == timebase, text
    data = (
        (800e-12, 2),
        (3.2e-9, 4),
        (6.4e-9, 5),
        (12.8e-9, 6),
        (19.2e-9, 7),
        (3.84e-8, 10),
        (6.144e-7, 100),
    )
    for (t, timebase) in data:
        text = "{} s does not fit timebase {}.".format(t, timebase)
        assert ps.getTimestepFromTimebase(timebase) == t, "Timestep: " + text
        try:
            timestep, _ = ps._lowLevelGetTimebase(timebase, 10, None, 0)
        except Exception:
            print(
                "getTimebase failed at time {}, timebase {}.".format(t,
                                                                     timebase))
            raise
        assert timestep == t, "LowLevel: {} != {}".format(timestep, t)
    print("Timebase test passed.")


def test_deviceResolution(ps):
    """Test setting/getting device resolution, including ADC limits."""
    ps.setResolution("12")
    assert ps.resolution == 1, "Picoscope variable was not set."
    assert ps.getResolution() == "12", "Picoscope resolution is wrong."
    assert ps.MIN_VALUE == -32736, "Minimum adc value is wrong."
    assert ps.MAX_VALUE == 32736, "Maximum adc value is wrong."
    print("Device resolution test passed.")


def test_rapid_block_mode(ps,
                          n_captures=100,
                          sample_interval=100e-9,  # 100 ns
                          sample_duration=2e-3,  # 1 ms
                          ):
    """Test the rapid block mode."""
    # Configuration of Picoscope
    ps.setChannel(channel="A", coupling="DC", VRange=1)
    ps.setChannel(channel="B", enabled=False)
    ps.setChannel(channel="C", enabled=False)
    ps.setChannel(channel="D", enabled=False)

    ps.setResolution('12')
    ps.setSamplingInterval(sample_interval, sample_duration)
    ps.setSimpleTrigger("A", threshold_V=0.1, timeout_ms=1)

    samples_per_segment = ps.memorySegments(n_captures)
    ps.setNoOfCaptures(n_captures)

    data = np.zeros((n_captures, samples_per_segment), dtype=np.int16)

    # Measurement
    t1 = time.time()

    ps.runBlock()
    ps.waitReady()

    t2 = time.time()
    print("Time to record data to scope: ", str(t2 - t1))

    # downSampleMode raw (no downsampling) is 0x80000000. 0 is invalid!
    ps.getDataRawBulk(data=data, downSampleMode=0x80000000)

    t3 = time.time()
    print("Time to copy to RAM: ", str(t3 - t2))

    plt.imshow(data[:, 0:ps.noSamples], aspect='auto', interpolation='none',
               cmap=plt.cm.hot)
    plt.colorbar()
    plt.title("rapid block mode")
    plt.show()
    print("Rapid block mode test passed.")


class Handler:

    def printing(self, text):
        print(text)

    def data_ready(self, handle, status, noOfSamples, overflow,
                   pParameter=None):
        """Show the asynchronously received data."""
        if status == 0:
            self.printing("{} samples received with overflow: {}".format(
                noOfSamples, overflow))
            plt.plot(self.data)
            plt.title("async")
            plt.show()
            self.ps._lowLevelClearDataBuffer(self.config[0], 0,
                                             downSampleMode=0x80000000)
            self.printing("Data reading asynchronously test passed.")
        else:
            self.printing("Data receiving error {}.".format(status))

    def test_read_async(self, handle=None, status=0, pParameter=None):
        """
        Test reading data asynchronously.

        If you call it manually instead of using it as a callback, use
        pParameter for handing over the picoscope instance.
        """
        if pParameter is not None:
            ps = pParameter
        else:
            ps = self.ps
        if status == 0:
            self.printing("Block is ready and can be read.")
            # config is a global variable written by the caller.
            channel, numSamples = self.config
            # Define data for data_ready.
            self.data = np.zeros(ps.noSamples, dtype=np.int16)
            if not isinstance(channel, int):
                channel = ps.CHANNELS[channel]
            self.config = channel, numSamples
            ps._lowLevelClearDataBufferAll(channel, 0)
            ps._lowLevelSetDataBuffer(channel, self.data,
                                      downSampleMode=0x80000000,
                                      segmentIndex=0)
            ps._lowLevelGetValuesAsync(ps.noSamples, 0, 1, 0x80000000, 0,
                                       self.data_ready, None)
            self.printing("Get values async started.")
        else:
            self.printing("Data is not ready. RunBlock had an error.")


def test_runBlock_async(ps, channel="A", sample_interval=100e-9,
                        sample_duration=2e-3):
    """Test running a block asynchronously."""
    # Define a handler to exchange data
    global handler
    handler = Handler()
    handler.ps = ps
    # Configuration of Picoscope
    ps.setChannel(channel=channel, coupling="DC", VRange=1)
    ps.memorySegments(1)
    ps.setNoOfCaptures(1)
    ps.setResolution('12')
    interval, samples, maxSamples = ps.setSamplingInterval(sample_interval,
                                                           sample_duration)
    ps.setSimpleTrigger("A", threshold_V=0.1, timeout_ms=1)

    handler.config = channel, samples
    # Run the block
    ps.runBlock(callback=handler.test_read_async)
    print("Run block started, waiting 1 s.")
    time.sleep(1)
    print("Run block finished.")


def test_downsampling(ps,
                      sample_interval=100e-9,  # 100 ns
                      sample_duration=2e-3,  # 1 ms
                      ):
    """Test for different downsampling methods."""
    ps._lowLevelClearDataBufferAll()
    ps.setChannel(channel="A", coupling="DC", VRange=1)
    ps.setChannel(channel="B", enabled=False)
    ps.setChannel(channel="C", enabled=False)
    ps.setChannel(channel="D", enabled=False)

    ps.setResolution('12')
    ps.memorySegments(1)
    ps.setNoOfCaptures(1)
    interval, samples, maxSamples = ps.setSamplingInterval(sample_interval,
                                                           sample_duration)
    ps.setSimpleTrigger("A", threshold_V=0.1, timeout_ms=1)

    ps.runBlock()
    ps.waitReady()

    data0 = np.zeros(ps.noSamples, dtype=np.int16)
    data1 = np.zeros(ps.noSamples, dtype=np.int16)
    data2 = np.zeros(ps.noSamples, dtype=np.int16)

    # downSampleMode raw (no downsampling) is 0x80000000. 0 is invalid!
    ps.getDataRaw(data=data0, downSampleMode=0x80000000)
    ps.getDataRaw(data=data1, downSampleMode=ps.RATIO_MODE['decimate'],
                  downSampleRatio=10)
    ps.getDataRaw(data=data2, downSampleMode=ps.RATIO_MODE['average'],
                  downSampleRatio=10)

    samplesReduced = len(data0) // 10
    plt.plot(data0, label="raw")
    plt.plot(range(0, 10 * samplesReduced, 10)[:samplesReduced],
             data1[:samplesReduced], label="decimate")
    plt.plot(range(0, 10 * samplesReduced, 10)[:samplesReduced],
             data2[:samplesReduced], label="average")
    plt.title("downsampling")
    plt.legend()
    plt.show()
    print("Downsampling test passed.")


if __name__ == "__main__":
    """Run all the tests."""
    # Initialize the picoscope
    ps = ps6000a.PS6000a()

    try:
        # Run tests.
        test_general_unit_calls(ps)
        test_timebase(ps)
        test_deviceResolution(ps)
        test_rapid_block_mode(ps)
        test_downsampling(ps)
        test_runBlock_async(ps)
    finally:
        # Close the connection
        ps.close()
    print("All tests passed.")



"""
This is the low level driver file for a specific Picoscope.

By this, I mean if parameters want to get passed as strings, they should be
handled by PSBase
All functions here should take things as close to integers as possible, the
only exception here is for array parameters. Array parameters should be passed
in a pythonic way through numpy since the PSBase class should not be aware of
the specifics behind how the clib is called.

The functions should not have any default values as these should be handled
by PSBase.
"""

# pico-python is Copyright (c) 2013-2014 By:
# Colin O'Flynn <coflynn@newae.com>
# Mark Harfouche <mark.harfouche@gmail.com>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import math

# to load the proper dll
import platform

# Do not import or use ill definied data types
# such as short int or long
# use the values specified in the h file
# float is always defined as 32 bits
# double is defined as 64 bits
from ctypes import byref, POINTER, create_string_buffer, c_float, \
    c_int16, c_int32, c_uint16, c_uint32, c_void_p
from ctypes import c_int32 as c_enum

from picoscope.picobase import _PicoscopeBase


class PS3000a(_PicoscopeBase):
    """The following are low-level functions for the PS3000a."""

    LIBNAME = "ps3000a"

    NUM_CHANNELS = 4
    CHANNELS = {"A": 0, "B": 1, "C": 2, "D": 3,
                "External": 4, "MaxChannels": 4, "TriggerAux": 5}

    ADC_RESOLUTIONS = {"8": 0, "12": 1, "14": 2, "15": 3, "16": 4}

    CHANNEL_RANGE = [{"rangeV": 10E-3, "apivalue": 0, "rangeStr": "10 mV"},
                     {"rangeV": 20E-3, "apivalue": 1, "rangeStr": "20 mV"},
                     {"rangeV": 50E-3, "apivalue": 2, "rangeStr": "50 mV"},
                     {"rangeV": 100E-3, "apivalue": 3, "rangeStr": "100 mV"},
                     {"rangeV": 200E-3, "apivalue": 4, "rangeStr": "200 mV"},
                     {"rangeV": 500E-3, "apivalue": 5, "rangeStr": "500 mV"},
                     {"rangeV": 1.0, "apivalue": 6, "rangeStr": "1 V"},
                     {"rangeV": 2.0, "apivalue": 7, "rangeStr": "2 V"},
                     {"rangeV": 5.0, "apivalue": 8, "rangeStr": "5 V"},
                     {"rangeV": 10.0, "apivalue": 9, "rangeStr": "10 V"},
                     {"rangeV": 20.0, "apivalue": 10, "rangeStr": "20 V"},
                     {"rangeV": 50.0, "apivalue": 11, "rangeStr": "50 V"},
                     ]

    CHANNEL_COUPLINGS = {"DC": 1, "AC": 0}

    # has_sig_gen = True
    WAVE_TYPES = {"Sine": 0, "Square": 1, "Triangle": 2,
                  "RampUp": 3, "RampDown": 4,
                  "Sinc": 5, "Gaussian": 6, "HalfSine": 7, "DCVoltage": 8,
                  "WhiteNoise": 9}

    SWEEP_TYPES = {"Up": 0, "Down": 1, "UpDown": 2, "DownUp": 3}

    SIGGEN_TRIGGER_TYPES = {"Rising": 0, "Falling": 1,
                            "GateHigh": 2, "GateLow": 3}
    SIGGEN_TRIGGER_SOURCES = {"None": 0, "ScopeTrig": 1, "AuxIn": 2,
                              "ExtIn": 3, "SoftTrig": 4, "TriggerRaw": 5}

    # This is actually different depending on the AB/CD models
    # I wonder how we could detect the difference between the oscilloscopes
    # I believe we can obtain this information from the setInfo function
    # by readign the hardware version
    # for the PS6403B version, the hardware version is "1 1",
    # an other possibility is that the PS6403B shows up as 6403 when using
    # VARIANT_INFO and others show up as PS6403X where X = A,C or D

    AWGPhaseAccumulatorSize = 32
    AWGBufferAddressWidth = 14
    AWGMaxSamples = 2 ** AWGBufferAddressWidth

    AWGDACInterval = 5E-9  # in seconds
    AWGDACFrequency = 1 / AWGDACInterval

    # Note this is NOT what is written in the Programming guide as of version
    # 10_5_0_28
    # This issue was acknowledged in this thread
    # http://www.picotech.com/support/topic13217.html
    AWGMaxVal = 32767
    AWGMinVal = -32767

    AWG_INDEX_MODES = {"Single": 0, "Dual": 1, "Quad": 2}

    MAX_VALUE_8BIT = 32512
    MIN_VALUE_8BIT = -32512
    MAX_VALUE_OTHER = 32767
    MIN_VALUE_OTHER = -32767

    EXT_RANGE_VOLTS = 5

    def __init__(self, serialNumber=None, connect=True):
        """Load DLL etc."""
        if platform.system() == 'Linux':
            from ctypes import cdll
            self.lib = cdll.LoadLibrary("lib" + self.LIBNAME + ".so")
        elif platform.system() == 'Darwin':
            from picoscope.darwin_utils import LoadLibraryDarwin
            self.lib = LoadLibraryDarwin("lib" + self.LIBNAME + ".dylib")
        else:
            from ctypes import windll
            from ctypes.util import find_library
            self.lib = windll.LoadLibrary(
                find_library(str(self.LIBNAME + ".dll"))
            )

        self.resolution = self.ADC_RESOLUTIONS["8"]

        super(PS3000a, self).__init__(serialNumber, connect)

    def _lowLevelOpenUnit(self, serialNumber):
        c_handle = c_int16()
        if serialNumber is not None:
            serialNumberStr = create_string_buffer(bytes(serialNumber,
                                                         encoding='utf-8'))
        else:
            serialNumberStr = None
        # Passing None is the same as passing NULL
        m = self.lib.ps3000aOpenUnit(byref(c_handle), serialNumberStr)
        self.handle = c_handle.value

        # copied over from ps5000a:
        # This will check if the power supply is not connected
        # and change the power supply accordingly
        # Personally (me = Mark), I don't like this
        # since the user should address this immediately, and we
        # shouldn't let this go as a soft error
        # but I think this should do for now
        if m == 0x11A:
            self.changePowerSource("PICO_POWER_SUPPLY_NOT_CONNECTED")
        else:
            self.checkResult(m)

    def _lowLevelCloseUnit(self):
        m = self.lib.ps3000aCloseUnit(c_int16(self.handle))
        self.checkResult(m)

    def _lowLevelEnumerateUnits(self):
        count = c_int16(0)
        m = self.lib.ps3000aEnumerateUnits(byref(count), None, None)
        self.checkResult(m)
        # a serial number is rouhgly 8 characters
        # an extra character for the comma
        # and an extra one for the space after the comma?
        # the extra two also work for the null termination
        serialLth = c_int16(count.value * (8 + 2))
        serials = create_string_buffer(serialLth.value + 1)

        m = self.lib.ps3000aEnumerateUnits(byref(count), serials,
                                           byref(serialLth))
        self.checkResult(m)

        serialList = str(serials.value.decode('utf-8')).split(',')

        serialList = [x.strip() for x in serialList]

        return serialList

    def _lowLevelSetChannel(self, chNum, enabled, coupling, VRange, VOffset,
                            BWLimited):
        m = self.lib.ps3000aSetChannel(c_int16(self.handle), c_enum(chNum),
                                       c_int16(enabled), c_enum(coupling),
                                       c_enum(VRange), c_float(VOffset))
        self.checkResult(m)

        m = self.lib.ps3000aSetBandwidthFilter(c_int16(self.handle),
                                               c_enum(chNum),
                                               c_enum(BWLimited))
        self.checkResult(m)

    def _lowLevelStop(self):
        m = self.lib.ps3000aStop(c_int16(self.handle))
        self.checkResult(m)

    def _lowLevelGetUnitInfo(self, info):
        s = create_string_buffer(256)
        requiredSize = c_int16(0)

        m = self.lib.ps3000aGetUnitInfo(c_int16(self.handle), byref(s),
                                        c_int16(len(s)), byref(requiredSize),
                                        c_enum(info))
        self.checkResult(m)
        if requiredSize.value > len(s):
            s = create_string_buffer(requiredSize.value + 1)
            m = self.lib.ps3000aGetUnitInfo(c_int16(self.handle), byref(s),
                                            c_int16(len(s)),
                                            byref(requiredSize), c_enum(info))
            self.checkResult(m)

        # should this bee ascii instead?
        # I think they are equivalent...
        return s.value.decode('utf-8')

    def _lowLevelFlashLed(self, times):
        m = self.lib.ps3000aFlashLed(c_int16(self.handle), c_int16(times))
        self.checkResult(m)

    def _lowLevelSetSimpleTrigger(self, enabled, trigsrc, threshold_adc,
                                  direction, delay, auto):
        m = self.lib.ps3000aSetSimpleTrigger(
            c_int16(self.handle), c_int16(enabled),
            c_enum(trigsrc), c_int16(threshold_adc),
            c_enum(direction), c_uint32(delay), c_int16(auto))
        self.checkResult(m)

    def _lowLevelSetNoOfCaptures(self, numCaptures):
        m = self.lib.ps3000aSetNoOfCaptures(c_int16(self.handle),
                                            c_uint16(numCaptures))
        self.checkResult(m)

    def _lowLevelMemorySegments(self, numSegments):
        maxSamples = c_int32()
        m = self.lib.ps3000aMemorySegments(c_int16(self.handle),
                                           c_uint16(numSegments),
                                           byref(maxSamples))
        self.checkResult(m)
        return maxSamples.value

    def _lowLevelGetMaxSegments(self):
        maxSegments = c_int16()
        m = self.lib.ps3000aGetMaxSegments(c_int16(self.handle),
                                           byref(maxSegments))
        self.checkResult(m)
        return maxSegments.value

    def _lowLevelRunBlock(self, numPreTrigSamples, numPostTrigSamples,
                          timebase, oversample, segmentIndex, callback,
                          pParameter):
        # NOT: Oversample is NOT used!
        timeIndisposedMs = c_int32()
        m = self.lib.ps3000aRunBlock(
            c_int16(self.handle), c_uint32(numPreTrigSamples),
            c_uint32(numPostTrigSamples), c_uint32(timebase),
            c_int16(oversample), byref(timeIndisposedMs),
            c_uint32(segmentIndex),
            c_void_p(), c_void_p())
        # According to the documentation, 'callback, pParameter' should work
        # instead of the last two c_void_p parameters.
        self.checkResult(m)
        return timeIndisposedMs.value

    def _lowLevelIsReady(self):
        ready = c_int16()
        m = self.lib.ps3000aIsReady(c_int16(self.handle), byref(ready))
        self.checkResult(m)
        if ready.value:
            return True
        else:
            return False

    def _lowlevelPingUnit(self):
        m = self.lib.ps3000aPingUnit(c_int16(self.handle))
        return m

    def _lowLevelGetTimebase(self, tb, noSamples, oversample, segmentIndex):
        """Return (timeIntervalSeconds, maxSamples)."""
        maxSamples = c_int32()
        intervalNanoSec = c_float()

        m = self.lib.ps3000aGetTimebase2(c_int16(self.handle), c_uint32(tb),
                                         c_uint32(noSamples),
                                         byref(intervalNanoSec),
                                         c_int16(oversample),
                                         byref(maxSamples),
                                         c_uint32(segmentIndex))
        self.checkResult(m)
        # divide by 1e9 to return interval in seconds
        return (intervalNanoSec.value * 1e-9, maxSamples.value)

    def getTimeBaseNum(self, sampleTimeS):
        """Convert sample time in S to something to pass to API Call."""
        maxSampleTime = (((2 ** 32 - 1) - 2) / 125000000)
        if sampleTimeS < 8.0E-9:
            st = math.floor(math.log(sampleTimeS * 1E9, 2))
            st = max(st, 0)
        else:
            if sampleTimeS > maxSampleTime:
                sampleTimeS = maxSampleTime
            st = math.floor((sampleTimeS * 125000000) + 2)

        # is this cast needed?
        st = int(st)
        return st

    def getTimestepFromTimebase(self, timebase):
        """Take API timestep code and returns the sampling period.

        API timestep is an integer from 0-32
        """
        if timebase < 3:
            dt = 2. ** timebase / 1.0E9
        else:
            dt = (timebase - 2.0) / 125000000.
        return dt

    def _lowLevelSetAWGSimpleDeltaPhase(self, waveform, deltaPhase,
                                        offsetVoltage, pkToPk, indexMode,
                                        shots, triggerType, triggerSource):
        """Waveform should be an array of shorts."""
        waveformPtr = waveform.ctypes.data_as(POINTER(c_int16))

        m = self.lib.ps3000aSetSigGenArbitrary(
            c_int16(self.handle),
            c_uint32(int(offsetVoltage * 1E6)),  # offset voltage in microvolts
            c_uint32(int(pkToPk * 1E6)),         # pkToPk in microvolts
            c_uint32(int(deltaPhase)),           # startDeltaPhase
            c_uint32(int(deltaPhase)),           # stopDeltaPhase
            c_uint32(0),                         # deltaPhaseIncrement
            c_uint32(0),                         # dwellCount
            waveformPtr,                         # arbitraryWaveform
            c_int32(len(waveform)),              # arbitraryWaveformSize
            c_enum(0),                           # sweepType for deltaPhase
            c_enum(0),            # operation (adding random noise and whatnot)
            c_enum(indexMode),                   # single, dual, quad
            c_uint32(shots),
            c_uint32(0),                         # sweeps
            c_uint32(triggerType),
            c_uint32(triggerSource),
            c_int16(0))                          # extInThreshold
        self.checkResult(m)

    def _lowLevelSetDataBuffer(self, channel, data, downSampleMode,
                               segmentIndex):
        """Set the data buffer.

        Be sure to call _lowLevelClearDataBuffer
        when you are done with the data array
        or else subsequent calls to GetValue will still use the same array.
        """
        dataPtr = data.ctypes.data_as(POINTER(c_int16))
        numSamples = len(data)

        m = self.lib.ps3000aSetDataBuffer(
            c_int16(self.handle), c_enum(channel),
            dataPtr, c_int32(numSamples),
            c_uint32(segmentIndex),
            c_enum(downSampleMode))
        self.checkResult(m)

    def _lowLevelSetDataBufferBulk(self, channel, data, segmentIndex,
                                   downSampleMode):
        """Set the bulk data buffer.

        In ps3000a, ps3000aSetDataBuffer combines the functionality of
        psX000YSetDataBuffer and psX000YSetDataBufferBulk. Since the rapid
        block functions in picoscope.py call the Bulk version, a delegator is
        needed. Note that the order of segmentIndex and downSampleMode is
        reversed.
        """
        self._lowLevelSetDataBuffer(channel, data, downSampleMode,
                                    segmentIndex)

    def _lowLevelSetMultipleDataBuffers(self, channel, data, downSampleMode):
        max_segments = self._lowLevelGetMaxSegments()
        if data.shape[0] < max_segments:
            raise ValueError("data array has fewer rows than current number " +
                             "of memory segments")
        if data.shape[1] < self.maxSamples:
            raise ValueError("data array has fewer columns than maxSamples")

        for i in range(max_segments):
            m = self._lowLevelSetDataBuffer(channel, data[i, :],
                                            downSampleMode, i)
            self.checkResult(m)

    def _lowLevelClearDataBuffer(self, channel, segmentIndex):
        m = self.lib.ps3000aSetDataBuffer(c_int16(self.handle),
                                          c_enum(channel),
                                          c_void_p(), c_uint32(0),
                                          c_uint32(segmentIndex),
                                          c_enum(0))
        self.checkResult(m)

    def _lowLevelGetValues(self, numSamples, startIndex, downSampleRatio,
                           downSampleMode, segmentIndex):
        numSamplesReturned = c_uint32()
        numSamplesReturned.value = numSamples
        overflow = c_int16()
        m = self.lib.ps3000aGetValues(
            c_int16(self.handle), c_uint32(startIndex),
            byref(numSamplesReturned), c_uint32(downSampleRatio),
            c_enum(downSampleMode), c_uint32(segmentIndex),
            byref(overflow))
        self.checkResult(m)
        return (numSamplesReturned.value, overflow.value)

    def _lowLevelGetValuesBulk(self, numSamples, fromSegment, toSegment,
                               downSampleRatio, downSampleMode, overflow):
        m = self.lib.ps3000aGetValuesBulk(
            c_int16(self.handle),
            byref(c_uint32(numSamples)),
            c_uint32(fromSegment),
            c_uint32(toSegment),
            c_uint32(downSampleRatio),
            c_int16(downSampleMode),
            overflow.ctypes.data_as(POINTER(c_int16))
            )
        self.checkResult(m)
        return overflow, numSamples

    def _lowLevelSetSigGenBuiltInSimple(self, offsetVoltage, pkToPk, waveType,
                                        frequency, shots, triggerType,
                                        triggerSource, stopFreq, increment,
                                        dwellTime, sweepType, numSweeps):
        # TODO, I just noticed that V2 exists
        # Maybe change to V2 in the future
        if stopFreq is None:
            stopFreq = frequency

        m = self.lib.ps3000aSetSigGenBuiltIn(
            c_int16(self.handle),
            c_int32(int(offsetVoltage * 1000000)),
            c_int32(int(pkToPk * 1000000)),
            c_int16(waveType),
            c_float(frequency), c_float(stopFreq),
            c_float(increment), c_float(dwellTime),
            c_enum(sweepType), c_enum(0),
            c_uint32(shots), c_uint32(numSweeps),
            c_enum(triggerType), c_enum(triggerSource),
            c_int16(0))
        self.checkResult(m)

    def _lowLevelChangePowerSource(self, powerstate):
        m = self.lib.ps3000aChangePowerSource(
            c_int16(self.handle),
            c_enum(powerstate))
        self.checkResult(m)

    def _lowLevelSigGenSoftwareControl(self, triggerType):
        m = self.lib.ps3000aSigGenSoftwareControl(
            c_int16(self.handle),
            c_enum(triggerType))
        self.checkResult(m)



from ctypes import cdll
from pathlib import Path
import shutil
import subprocess
import tempfile

import sys


def LoadLibraryDarwin(library):
    """Wrapper around cdll.LoadLibrary that works around how SIP breaks
    dynamically linked libraries. This improves upon the process described
    here:

    http://ulthiel.com/vk2utl/picoscope-python-interface-under-mac-os-x/
    """
    PICO_LIB_PATH = "/Applications/PicoScope6.app/Contents/Resources/lib/"

    # Libraries that depend on libiomp5.dylib
    IOMP5_DEPS = ["libpicoipp.dylib", "libpicoipp.1.dylib"]

    try:
        # Get the library normally. This should work if SIP is disabled and
        # DYLD_LIBRARY_PATH is set properly.
        return cdll.LoadLibrary(library)
    except OSError:
        # 2.7 Fix. This only fixes flake8
        if sys.version_info[0] == 2:
            FileNotFoundError = IOError
        if not Path(PICO_LIB_PATH).is_dir():
            raise FileNotFoundError(
                    "/Applications/PicoScope6.app is missing")

        # Modifying the libraries in-place breaks their signatures, causing
        # PicoScope6.app to fail to detect the oscilloscope.  Instead, patch
        # copies of the libraries in a temporary directory.
        tempLibDir = tempfile.TemporaryDirectory()
        patchedPicoPath = tempLibDir.name + "/lib/"
        shutil.copytree(PICO_LIB_PATH, patchedPicoPath)

        # Patch libraries that depend on libiomp5.dylib to refer to it by an
        # absolute path instead of a relative path, which is not allowed by
        # SIP.
        for libraryToPatch in IOMP5_DEPS:
            subprocess.run(["install_name_tool", "-change", "libiomp5.dylib",
                            patchedPicoPath + "/libiomp5.dylib",
                            patchedPicoPath + "/" + libraryToPatch
                            ]).check_returncode()

        # Finally, patch the originally requested library to look in the
        # patched directory.
        patchedLibrary = patchedPicoPath + "/" + library
        subprocess.run(["install_name_tool", "-add_rpath", patchedPicoPath,
                        patchedLibrary]).check_returncode()
        loadedLibrary = cdll.LoadLibrary(patchedLibrary)

        # Sneak the directory into the library so it's not deleted until the
        # library is garbage collected.
        loadedLibrary.tempLibDir = tempLibDir
        return loadedLibrary



# This is the instrument-specific file for the PS4000A series of instruments.
#
# pico-python is Copyright (c) 2013-2014 By:
# Colin O'Flynn <coflynn@newae.com>
# Mark Harfouche <mark.harfouche@gmail.com>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
This is the low level driver file for a specific Picoscope.

By this, I mean if parameters want to get passed as strings, they should be
handled by PSBase
All functions here should take things as close to integers as possible, the
only exception here is for array parameters. Array parameters should be passed
in a pythonic way through numpy since the PSBase class should not be aware of
the specifics behind how the clib is called.

The functions should not have any default values as these should be handled
by PSBase.

"""

# 3.0 compatibility
# see http://docs.python.org/2/library/__future__.html
from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import math

# to load the proper dll
import platform

# Do not import or use ill definied data types
# such as short int or long
# use the values specified in the h file
# float is always defined as 32 bits
# double is defined as 64 bits
from ctypes import byref, POINTER, create_string_buffer, c_float, c_double, \
    c_int16, c_uint16, c_int32, c_uint32, c_uint64, c_void_p, c_int8, \
    CFUNCTYPE
from ctypes import c_int32 as c_enum

from picoscope.picobase import _PicoscopeBase


# Decorators for callback functions. PICO_STATUS is uint32_t.
def blockReady(function):
    """typedef void (*ps4000aBlockReady)
    (
     int16_t         handle,
     PICO_STATUS     status,
     void          * pParameter
    )
    """
    if function is None:
        return None

    callback = CFUNCTYPE(c_void_p, c_int16, c_uint32, c_void_p)
    return callback(function)


def dataReady(function):
    """typedef void (*ps4000aDataReady)
    (
     int16_t         handle,
     PICO_STATUS     status,
     uint32_t        noOfSamples,
     int16_t         overflow,
     void         * pParameter
    )
    """
    callback = CFUNCTYPE(c_void_p,
                         c_int16, c_uint32, c_uint32, c_int16, c_void_p)
    return callback(function)


def streamingReady(function):
    """typedef void (*ps4000aStreamingReady)
    (
        int16_t     handle,
        int32_t     noOfSamples,
        uint32_t    startIndex,
        int16_t     overflow,
        uint32_t    triggerAt,
        int16_t     triggered,
        int16_t     autoStop,
        void      * pParameter
    )
    """
    callback = CFUNCTYPE(c_void_p, c_int16, c_int32, c_uint32, c_int16,
                         c_uint32, c_int16, c_int16, c_void_p)
    return callback(function)


class PS4000a(_PicoscopeBase):
    """The following are low-level functions for the PS4000A."""

    LIBNAME = "ps4000a"

    MAX_VALUE = 32764
    MIN_VALUE = -32764

    # EXT/AUX seems to have an input impedance of 50 ohm (PS6403B)
    EXT_MAX_VALUE = 32767
    EXT_MIN_VALUE = -32767
    EXT_RANGE_VOLTS = 1

    # I don't think that the 50V range is allowed, but I left it there anyway
    # The 10V and 20V ranges are only allowed in high impedence modes
    CHANNEL_RANGE = [{"rangeV": 20E-3, "apivalue": 1, "rangeStr": "20 mV"},
                     {"rangeV": 50E-3, "apivalue": 2, "rangeStr": "50 mV"},
                     {"rangeV": 100E-3, "apivalue": 3, "rangeStr": "100 mV"},
                     {"rangeV": 200E-3, "apivalue": 4, "rangeStr": "200 mV"},
                     {"rangeV": 500E-3, "apivalue": 5, "rangeStr": "500 mV"},
                     {"rangeV": 1.0, "apivalue": 6, "rangeStr": "1 V"},
                     {"rangeV": 2.0, "apivalue": 7, "rangeStr": "2 V"},
                     {"rangeV": 5.0, "apivalue": 8, "rangeStr": "5 V"},
                     {"rangeV": 10.0, "apivalue": 9, "rangeStr": "10 V"},
                     {"rangeV": 20.0, "apivalue": 10, "rangeStr": "20 V"},
                     {"rangeV": 50.0, "apivalue": 11, "rangeStr": "50 V"},
                     ]

    NUM_CHANNELS = 8
    CHANNELS = {"A": 0, "B": 1, "C": 2, "D": 3, "E": 4, "F": 5, "G": 6,
                "H": 7,  "MaxChannels": 8}

    ADC_RESOLUTIONS = {"8": 0, "12": 1, "14": 2, "15": 3, "16": 4}

    CHANNEL_COUPLINGS = {"DC50": 2, "DC": 1, "AC": 0}

    WAVE_TYPES = {"Sine": 0, "Square": 1, "Triangle": 2,
                  "RampUp": 3, "RampDown": 4,
                  "Sinc": 5, "Gaussian": 6, "HalfSine": 7, "DCVoltage": 8,
                  "WhiteNoise": 9}

    SWEEP_TYPES = {"Up": 0, "Down": 1, "UpDown": 2, "DownUp": 3}

    TIME_UNITS = {"femtoseconds": 0,
                  "picoseconds": 1,
                  "nanoseconds": 2,
                  "microseconds": 3,
                  "milliseconds": 4,
                  "seconds": 5}

    SIGGEN_TRIGGER_TYPES = {"Rising": 0, "Falling": 1,
                            "GateHigh": 2, "GateLow": 3}

    SIGGEN_TRIGGER_SOURCES = {"None": 0, "ScopeTrig": 1,
                              "AuxIn": 2, "ExtIn": 3, "SoftTrig": 4}

    AWGPhaseAccumulatorSize = 32
    AWGBufferAddressWidth = 14
    AWGMaxSamples = 2 ** AWGBufferAddressWidth

    AWGDACInterval = 12.5E-9  # in seconds
    AWGDACFrequency = 1 / AWGDACInterval

    # From the programmer's guide, p.99, defined for the PicoScope 4824. Values
    # have been checked against those returned by the
    # ps4000aSigGenArbitraryMinMaxValues function, with a PS4824 device
    # connected.
    AWGMaxVal = 32767
    AWGMinVal = -32768

    AWG_INDEX_MODES = {"Single": 0, "Dual": 1, "Quad": 2}

    def __init__(self, serialNumber=None, connect=True):
        """Load DLLs."""
        self.handle = None

        if platform.system() == 'Linux':
            from ctypes import cdll
            # ok I don't know what is wrong with my installer,
            # but I need to include .so.2
            self.lib = cdll.LoadLibrary("lib" + self.LIBNAME + ".so.2")
        elif platform.system() == 'Darwin':
            from picoscope.darwin_utils import LoadLibraryDarwin
            self.lib = LoadLibraryDarwin("lib" + self.LIBNAME + ".dylib")
        else:
            from ctypes import windll
            from ctypes.util import find_library
            self.lib = windll.LoadLibrary(
                find_library(str(self.LIBNAME + ".dll"))
            )

        self.resolution = self.ADC_RESOLUTIONS["12"]

        super(PS4000a, self).__init__(serialNumber, connect)

    def _lowLevelOpenUnit(self, serialNumber):
        c_handle = c_int16()
        if serialNumber is not None:
            serialNumberStr = create_string_buffer(bytes(serialNumber,
                                                         encoding='utf-8'))
        else:
            serialNumberStr = None
        # Passing None is the same as passing NULL
        m = self.lib.ps4000aOpenUnit(byref(c_handle), serialNumberStr)
        self.handle = c_handle.value

        # This will check if the power supply is not connected
        # and change the power supply accordingly
        # Personally (me = Mark), I don't like this
        # since the user should address this immediately, and we
        # shouldn't let this go as a soft error
        # but I think this should do for now
        if m == 0x11A:
            self.changePowerSource(m)
        else:
            self.checkResult(m)

        self.model = self.getUnitInfo('VariantInfo')

    def _lowLevelOpenUnitAsync(self, serialNumber):
        c_status = c_int16()
        if serialNumber is not None:
            serialNumberStr = create_string_buffer(bytes(serialNumber,
                                                         encoding='utf-8'))
        else:
            serialNumberStr = None
        # Passing None is the same as passing NULL
        m = self.lib.ps4000aOpenUnitAsync(byref(c_status), serialNumberStr)
        self.checkResult(m)

        # Set the model after completion in _lowLevelOpenUnitProgress.
        return c_status.value

    def _lowLevelOpenUnitProgress(self):
        complete = c_int16()
        progressPercent = c_int16()
        handle = c_int16()

        m = self.lib.ps4000aOpenUnitProgress(byref(handle),
                                             byref(progressPercent),
                                             byref(complete))
        self.checkResult(m)

        if complete.value != 0:
            self.handle = handle.value
            self.model = self.getUnitInfo('VariantInfo')

        # if we only wanted to return one value, we could do somethign like
        # progressPercent = progressPercent * (1 - 0.1 * complete)
        return (progressPercent.value, complete.value)

    def _lowLevelCloseUnit(self):
        m = self.lib.ps4000aCloseUnit(c_int16(self.handle))
        self.checkResult(m)

    def _lowLevelEnumerateUnits(self):
        count = c_int16(0)
        serials = c_int8(0)
        serialLth = c_int16(0)

        m = self.lib.ps4000aEnumerateUnits(byref(count), byref(serials),
                                           byref(serialLth))
        self.checkResult(m)
        # a serial number is rouhgly 10 characters
        # an extra character for the comma
        # and an extra one for the space after the comma?
        # the extra two also work for the null termination
        serialLth = c_int16(count.value * (10 + 2))
        serials = create_string_buffer(serialLth.value + 1)

        m = self.lib.ps4000aEnumerateUnits(byref(count), serials,
                                           byref(serialLth))
        self.checkResult(m)

        serialList = str(serials.value.decode('utf-8')).split(',')

        serialList = [x.strip() for x in serialList]

        return serialList

    def _lowLevelSetChannel(self, chNum, enabled, coupling, VRange, VOffset,
                            BWLimited):
        m = self.lib.ps4000aSetChannel(c_int16(self.handle), c_enum(chNum),
                                       c_int16(enabled), c_enum(coupling),
                                       c_enum(VRange), c_float(VOffset))
        self.checkResult(m)
        # Only for model PS4444
        # See discussion: https://github.com/colinoflynn/pico-python/pull/171
        if self.model.startswith('4444'):  # Only for model 4444
            m = self.lib.ps4000aSetBandwidthFilter(c_int16(self.handle),
                                                   c_enum(chNum),
                                                   c_enum(BWLimited))
            self.checkResult(m)

    def _lowLevelStop(self):
        m = self.lib.ps4000aStop(c_int16(self.handle))
        self.checkResult(m)

    def _lowLevelGetUnitInfo(self, info):
        s = create_string_buffer(256)
        requiredSize = c_int16(0)

        m = self.lib.ps4000aGetUnitInfo(c_int16(self.handle), byref(s),
                                        c_int16(len(s)), byref(requiredSize),
                                        c_enum(info))
        self.checkResult(m)
        if requiredSize.value > len(s):
            s = create_string_buffer(requiredSize.value + 1)
            m = self.lib.ps4000aGetUnitInfo(c_int16(self.handle), byref(s),
                                            c_int16(len(s)),
                                            byref(requiredSize), c_enum(info))
            self.checkResult(m)

        # should this bee ascii instead?
        # I think they are equivalent...
        return s.value.decode('utf-8')

    def _lowLevelFlashLed(self, times):
        m = self.lib.ps4000aFlashLed(c_int16(self.handle), c_int16(times))
        self.checkResult(m)

    def _lowLevelSetSimpleTrigger(self, enabled, trigsrc, threshold_adc,
                                  direction, delay, timeout_ms):
        m = self.lib.ps4000aSetSimpleTrigger(
            c_int16(self.handle), c_int16(enabled),
            c_enum(trigsrc), c_int16(threshold_adc),
            c_enum(direction), c_uint32(delay), c_int16(timeout_ms))
        self.checkResult(m)

    def _lowLevelRunBlock(self, numPreTrigSamples, numPostTrigSamples,
                          timebase, oversample, segmentIndex, callback,
                          pParameter):
        # Hold a reference to the callback so that the Python
        # function pointer doesn't get free'd
        self._c_runBlock_callback = blockReady(callback)
        timeIndisposedMs = c_int32()
        m = self.lib.ps4000aRunBlock(
            c_int16(self.handle), c_int32(numPreTrigSamples),
            c_int32(numPostTrigSamples), c_uint32(timebase),
            byref(timeIndisposedMs), c_uint32(segmentIndex),
            self._c_runBlock_callback, c_void_p())
        self.checkResult(m)
        return timeIndisposedMs.value

    def _lowLevelIsReady(self):
        ready = c_int16()
        m = self.lib.ps4000aIsReady(c_int16(self.handle), byref(ready))
        self.checkResult(m)
        if ready.value:
            return True
        else:
            return False

    def _lowLevelGetTimebase(self, timebase, noSamples, oversample,
                             segmentIndex):
        """
        Calculate the sampling interval and maximum number of samples.

        timebase
            Number of the selected timebase.
        noSamples
            Number of required samples.
        oversample
            Not used.
        segmentIndex
            Index of the segment to save samples in

        Return
        -------
        timeIntervalSeconds : float
            Time interval between two samples in s.
        maxSamples : int
            maximum number of samples available depending on channels
            and timebase chosen.
        """
        maxSamples = c_int32()
        timeIntervalSeconds = c_float()

        m = self.lib.ps4000aGetTimebase2(c_int16(self.handle),
                                         c_uint32(timebase),
                                         c_int32(noSamples),
                                         byref(timeIntervalSeconds),
                                         byref(maxSamples),
                                         c_uint32(segmentIndex))
        self.checkResult(m)

        return (timeIntervalSeconds.value / 1.0E9, maxSamples.value)

    def getTimeBaseNum(self, sampleTimeS):
        """
        Convert `sampleTimeS` in s to the integer timebase number.

        See "Timebases" section of the PS4000a programmer's guide
        for more information.
        """
        if self.model == '4444':
            maxSampleTime = (((2 ** 32 - 1) - 2) / 5.0E7)

            if (sampleTimeS <= 2.5E-9 and
                    self.resolution == self.ADC_RESOLUTIONS["12"]):
                timebase = 0
            elif (sampleTimeS <= 20E-9 and
                    self.resolution == self.ADC_RESOLUTIONS["14"]):
                timebase = 3
            else:
                # Otherwise in range 2^32-1
                if sampleTimeS > maxSampleTime:
                    sampleTimeS = maxSampleTime

                timebase = math.floor((sampleTimeS * 5.0E7) + 2)

        else:
            timebase = math.floor(sampleTimeS / 12.5e-9 - 1)
            timebase = max(timebase, 0)
            timebase = min(timebase, 2 ** 32 - 1)

        return timebase

    def getTimestepFromTimebase(self, timebase):
        """
        Convert `timebase` index to sampletime in seconds.

        See "Timebases" section of the PS4000a programmer's guide
        for more information.
        """
        if self.model == '4444':
            if timebase <= 3:
                dt = 2 ** timebase / 4.0E8
            else:
                dt = (timebase - 2) / 5.0E7

        else:
            dt = (timebase + 1) / 8.0E7
        return dt

    def _lowLevelSetAWGSimpleDeltaPhase(self, waveform, deltaPhase,
                                        offsetVoltage, pkToPk, indexMode,
                                        shots, triggerType, triggerSource):
        """Waveform should be an array of shorts."""
        waveformPtr = waveform.ctypes.data_as(POINTER(c_int16))

        m = self.lib.ps4000aSetSigGenArbitrary(
            c_int16(self.handle),
            c_int32(int(offsetVoltage * 1E6)),  # offset voltage in microvolts
            c_uint32(int(pkToPk * 1E6)),         # pkToPk in microvolts
            c_uint32(int(deltaPhase)),           # startDeltaPhase
            c_uint32(int(deltaPhase)),           # stopDeltaPhase
            c_uint32(0),                         # deltaPhaseIncrement
            c_uint32(0),                         # dwellCount
            waveformPtr,                         # arbitraryWaveform
            c_int32(len(waveform)),              # arbitraryWaveformSize
            c_enum(0),                           # sweepType for deltaPhase
            c_enum(0),            # operation (adding random noise and whatnot)
            c_enum(indexMode),                   # single, dual, quad
            c_uint32(shots),
            c_uint32(0),                         # sweeps
            c_uint32(triggerType),
            c_uint32(triggerSource),
            c_int16(0))                          # extInThreshold
        self.checkResult(m)

    def _lowLevelSetDataBuffer(self, channel, data, downSampleMode,
                               segmentIndex):
        """Set the data buffer.

        Be sure to call _lowLevelClearDataBuffer
        when you are done with the data array
        or else subsequent calls to GetValue will still use the same array.
        """
        dataPtr = data.ctypes.data_as(POINTER(c_int16))
        numSamples = len(data)

        m = self.lib.ps4000aSetDataBuffer(c_int16(self.handle),
                                          c_enum(channel),
                                          dataPtr, c_uint32(numSamples),
                                          c_uint32(segmentIndex),
                                          c_uint32(downSampleMode))
        self.checkResult(m)

    def _lowLevelClearDataBuffer(self, channel, segmentIndex):
        m = self.lib.ps4000aSetDataBuffer(c_int16(self.handle),
                                          c_enum(channel),
                                          c_void_p(), c_uint32(0), c_uint32(0),
                                          c_enum(0))
        self.checkResult(m)

    def _lowLevelGetValues(self, numSamples, startIndex, downSampleRatio,
                           downSampleMode, segmentIndex):
        numSamplesReturned = c_uint32()
        numSamplesReturned.value = numSamples
        overflow = c_int16()
        m = self.lib.ps4000aGetValues(
            c_int16(self.handle), c_uint32(startIndex),
            byref(numSamplesReturned), c_uint32(downSampleRatio),
            c_enum(downSampleMode), c_uint16(segmentIndex),
            byref(overflow))
        self.checkResult(m)
        return (numSamplesReturned.value, overflow.value)

    def _lowLevelGetValuesAsync(self, numSamples, startIndex, downSampleRatio,
                                downSampleMode, segmentIndex, callback, pPar):
        self._c_getValues_callback = dataReady(callback)
        m = self.lib.ps4000aGetValuesAsync(c_int16(self.handle),
                                           c_uint32(startIndex),
                                           c_uint32(numSamples),
                                           c_uint32(downSampleRatio),
                                           c_enum(downSampleMode),
                                           c_uint32(segmentIndex),
                                           self._c_getValues_callback,
                                           c_void_p())
        self.checkResult(m)

    def _lowLevelSetDeviceResolution(self, resolution):
        self.resolution = resolution
        m = self.lib.ps4000aSetDeviceResolution(
            c_int16(self.handle),
            c_enum(resolution))
        self.checkResult(m)

    def _lowLevelChangePowerSource(self, powerstate):
        m = self.lib.ps4000aChangePowerSource(
            c_int16(self.handle),
            c_enum(powerstate))
        self.checkResult(m)

    def _lowLevelGetValuesBulk(self, numSamples, fromSegment, toSegment,
                               downSampleRatio, downSampleMode, overflow):
        """Copy data from several memory segments at once."""
        overflowPoint = overflow.ctypes.data_as(POINTER(c_int16))
        m = self.lib.ps4000aGetValuesBulk(
            c_int16(self.handle),
            byref(c_int32(numSamples)),
            c_int32(fromSegment),
            c_int32(toSegment),
            c_int32(downSampleRatio),
            c_enum(downSampleMode),
            overflowPoint
        )
        self.checkResult(m)

    def _lowLevelSetDataBufferBulk(self, channel, data, segmentIndex,
                                   downSampleMode):
        """Just calls setDataBuffer with argument order changed.

        For compatibility with current picobase.py.
        """
        self._lowLevelSetDataBuffer(channel, data, downSampleMode,
                                    segmentIndex)

    def _lowLevelPingUnit(self):
        """Check connection to picoscope and return the error."""
        return self.lib.ps4000aPingUnit(c_int16(self.handle))

    def _lowLevelSetSigGenBuiltInSimple(self, offsetVoltage, pkToPk, waveType,
                                        frequency, shots, triggerType,
                                        triggerSource, stopFreq, increment,
                                        dwellTime, sweepType, numSweeps):
        if stopFreq is None:
            stopFreq = frequency

        m = self.lib.ps4000aSetSigGenBuiltIn(
            c_int16(self.handle),
            c_int32(int(offsetVoltage * 1000000)),
            c_uint32(int(pkToPk * 1000000)),
            c_enum(waveType),
            c_double(frequency), c_double(stopFreq),
            c_double(increment), c_double(dwellTime),
            c_enum(sweepType), c_enum(0),
            c_uint32(shots), c_uint32(numSweeps),
            c_enum(triggerType), c_enum(triggerSource),
            c_int16(0))
        self.checkResult(m)

    def _lowLevelSigGenSoftwareControl(self, state):
        m = self.lib.ps4000aSigGenSoftwareControl(
            c_int16(self.handle),
            c_int16(state))
        self.checkResult(m)

    ####################################################################
    # Untested functions below                                         #
    #                                                                  #
    ####################################################################

    def _lowLevelGetMaxDownSampleRatio(self, noOfUnaggregatedSamples,
                                       downSampleRatioMode, segmentIndex):
        maxDownSampleRatio = c_uint32()

        m = self.lib.ps4000aGetMaxDownSampleRatio(
            c_int16(self.handle),
            c_uint32(noOfUnaggregatedSamples),
            byref(maxDownSampleRatio),
            c_enum(downSampleRatioMode),
            c_uint16(segmentIndex))
        self.checkResult(m)

        return maxDownSampleRatio.value

    def _lowLevelGetNoOfCaptures(self):
        nCaptures = c_uint32()

        m = self.lib.ps4000aGetNoOfCaptures(c_int16(self.handle),
                                            byref(nCaptures))
        self.checkResult(m)

        return nCaptures.value

    def _lowLevelGetTriggerTimeOffset(self, segmentIndex):
        time = c_uint64()
        timeUnits = c_enum()

        m = self.lib.ps4000aGetTriggerTimeOffset64(
            c_int16(self.handle),
            byref(time),
            byref(timeUnits),
            c_uint16(segmentIndex))

        self.checkResult(m)

        if timeUnits.value == 0:  # PS4000a_FS
            return time.value * 1E-15
        elif timeUnits.value == 1:  # PS4000a_PS
            return time.value * 1E-12
        elif timeUnits.value == 2:  # PS4000a_NS
            return time.value * 1E-9
        elif timeUnits.value == 3:  # PS4000a_US
            return time.value * 1E-6
        elif timeUnits.value == 4:  # PS4000a_MS
            return time.value * 1E-3
        elif timeUnits.value == 5:  # PS4000a_S
            return time.value * 1E0
        else:
            raise TypeError("Unknown timeUnits %d" % timeUnits.value)

    def _lowLevelMemorySegments(self, nSegments):
        nMaxSamples = c_uint32()

        m = self.lib.ps4000aMemorySegments(c_int16(self.handle),
                                           c_uint16(nSegments),
                                           byref(nMaxSamples))
        self.checkResult(m)

        return nMaxSamples.value

    def _lowLevelSetDataBuffers(self, channel, bufferMax, bufferMin,
                                downSampleRatioMode):
        bufferMaxPtr = bufferMax.ctypes.data_as(POINTER(c_int16))
        bufferMinPtr = bufferMin.ctypes.data_as(POINTER(c_int16))
        bufferLth = len(bufferMax)

        m = self.lib.ps4000aSetDataBuffers(
            c_int16(self.handle),
            c_enum(channel),
            bufferMaxPtr,
            bufferMinPtr,
            c_uint32(bufferLth))
        self.checkResult(m)

    def _lowLevelClearDataBuffers(self, channel):
        m = self.lib.ps4000aSetDataBuffers(
            c_int16(self.handle),
            c_enum(channel),
            c_void_p(),
            c_void_p(),
            c_uint32(0))
        self.checkResult(m)

    def _lowLevelSetNoOfCaptures(self, nCaptures):
        m = self.lib.ps4000aSetNoOfCaptures(
            c_int16(self.handle),
            c_uint16(nCaptures))
        self.checkResult(m)

    # ETS Functions
    def _lowLevelSetEts(self):
        pass

    def _lowLevelSetEtsTimeBuffer(self):
        pass

    def _lowLevelSetEtsTimeBuffers(self):
        pass

    def _lowLevelSetExternalClock(self):
        pass

    # Complicated triggering
    # need to understand structs for this one to work
    def _lowLevelIsTriggerOrPulseWidthQualifierEnabled(self):
        pass

    def _lowLevelGetValuesTriggerTimeOffsetBulk(self):
        pass

    def _lowLevelSetTriggerChannelConditions(self):
        pass

    def _lowLevelSetTriggerChannelDirections(self):
        pass

    def _lowLevelSetTriggerChannelProperties(self):
        pass

    def _lowLevelSetPulseWidthQualifier(self):
        pass

    def _lowLevelSetTriggerDelay(self):
        pass

    # Async functions
    def _lowLevelGetValuesBulkAsync(self):
        pass

    # overlapped functions
    def _lowLevelGetValuesOverlapped(self):
        pass

    def _lowLevelGetValuesOverlappedBulk(self):
        pass

    # Streaming related functions
    def _lowLevelGetStreamingLatestValues(self, lpPs4000Ready,
                                          pParameter=c_void_p()):
        m = self.lib.ps4000aGetStreamingLatestValues(
            c_uint16(self.handle),
            lpPs4000Ready,
            pParameter)
        self.checkResult(m)

    def _lowLevelNoOfStreamingValues(self):
        noOfValues = c_uint32()

        m = self.lib.ps4000aNoOfStreamingValues(c_int16(self.handle),
                                                byref(noOfValues))
        self.checkResult(m)

        return noOfValues.value

    def _lowLevelRunStreaming(self, sampleInterval, sampleIntervalTimeUnits,
                              maxPreTriggerSamples, maxPostTriggerSamples,
                              autoStop, downSampleRatio, downSampleRatioMode,
                              overviewBufferSize):
        m = self.lib.ps4000aRunStreaming(
            c_int16(self.handle),
            byref(c_uint32(sampleInterval)),
            c_enum(sampleIntervalTimeUnits),
            c_uint32(maxPreTriggerSamples),
            c_uint32(maxPostTriggerSamples),
            c_int16(autoStop),
            c_uint32(downSampleRatio),
            c_enum(downSampleRatioMode),
            c_uint32(overviewBufferSize))

        self.checkResult(m)

    def _lowLevelStreamingReady(self):
        pass



# -*- coding: utf-8 -*-
"""
Testing the ps4000a Series.
"""

# Imports
import matplotlib.pyplot as plt
import numpy as np
from picoscope import ps4000a
import time


def test_general_unit_calls(picoscope):
    """Test general unit calls."""
    picoscope.flashLed()
    print(picoscope.getAllUnitInfo())
    assert not picoscope.ping(), "Ping failed."
    print("General unit calls test passed.")


def test_timebase(ps):
    """Test the timebase methods."""
    # Preparation
    ps.memorySegments(1)
    # Tests
    if ps.model == "4444":
        data = ((20e-9, 3),
                (40e-9, 4))
    else:
        data = ((25e-9, 1),
                (100e-9, 7))
    for (t, timebase) in data:
        text = f"time {t} does not fit timebase {timebase}."
        assert ps.getTimeBaseNum(t) == timebase, "timebasenum: " + text
        assert ps.getTimestepFromTimebase(timebase) == t, "Timestep " + text
        timestep, _ = ps._lowLevelGetTimebase(timebase, 10, None, 0)
        assert timestep == t, f"lowLevel: {timestep} != {t}"
    print("Timebase test passed.")


def test_deviceResolution(ps):
    """Test setting/getting device resolution."""
    if ps.model == "4444":
        ps.setResolution("12")
        assert ps.resolution == "12", "Resolution was not set."
        # assert ps.getResolution() == "12"  not implemented yet
        print("Device resolution test passed.")
    else:
        print("Model does not support resolution.")


def test_rapid_block_mode(ps,
                          n_captures=100,
                          sample_interval=100e-9,  # 100 ns
                          sample_duration=2e-3,  # 1 ms
                          ):
    """Test the rapid block mode."""
    # Configuration of Picoscope
    ps.setChannel(channel="A", coupling="DC", VRange=1)
    ps.setChannel(channel="B", enabled=False)

    ps.setSamplingInterval(sample_interval, sample_duration)
    ps.setSimpleTrigger("A", threshold_V=0.1, timeout_ms=1)

    samples_per_segment = ps.memorySegments(n_captures)
    ps.setNoOfCaptures(n_captures)

    data = np.zeros((n_captures, samples_per_segment), dtype=np.int16)

    # Measurement
    t1 = time.time()

    ps.runBlock()
    ps.waitReady()

    t2 = time.time()
    print("Time to record data to scope: ", str(t2 - t1))

    ps.getDataRawBulk(data=data)

    t3 = time.time()
    print("Time to copy to RAM: ", str(t3 - t2))

    plt.imshow(data[:, 0:ps.noSamples], aspect='auto', interpolation='none',
               cmap=plt.cm.hot)
    plt.colorbar()
    plt.show()
    print("Rapid block mode test passed.")


def data_ready(handle, status, noOfSamples, overflow, pParameter=None):
    """Show the asynchronously received data."""
    if status == 0:
        print(f"{noOfSamples} samples received with overflow: {overflow}")
        plt.plot(data)
        plt.show()
        print("Data reading asynchronously test passed.")
    else:
        print(f"Data receiving error {status}.")


def test_read_async(handle=None, status=0, pParameter=None):
    """Test reading data asynchronously."""
    if status == 0:
        print("Block is ready and can be read.")
        channel, numSamples = config
        global data
        data = np.empty(numSamples, dtype=np.int16)
        if not isinstance(channel, int):
            channel = ps.CHANNELS[channel]
        ps._lowLevelSetDataBuffer(channel, data, 0, 0)
        ps._lowLevelGetValuesAsync(numSamples, 0, 1, 0, 0, data_ready, None)
        print("Get values async started.")
    else:
        print("Data is not ready. RunBlock had an error.")


def test_runBlock_async(picoscope, channel="A", sample_interval=100e-9,
                        sample_duration=2e-3):
    """Test running a block asynchronously."""
    # Configuration of Picoscope
    global ps
    ps = picoscope
    ps.setChannel(channel=channel, coupling="DC", VRange=1)
    ps.memorySegments(1)
    ps.setNoOfCaptures(1)
    i, samples, m = ps.setSamplingInterval(sample_interval, sample_duration)
    ps.setSimpleTrigger("A", threshold_V=0.1, timeout_ms=1)

    global config
    config = channel, samples
    # Run the block
    ps.runBlock(callback=test_read_async)
    print("Run Block started, waiting 2 s.")
    time.sleep(2)
    print("Run block finished")


if __name__ == "__main__":
    """Run all the tests."""
    # Initialize the picoscope
    ps = ps4000a.PS4000a()

    try:
        # Run tests.
        test_general_unit_calls(ps)
        test_deviceResolution(ps)
        test_rapid_block_mode(ps)
        test_runBlock_async(ps)
    finally:
        # Close the connection
        ps.close()
    print("All tests passed.")



# -*- coding: utf-8 -*-
"""
This is the base class that all picoscope modules use.

As much logic as possible is put into this file.
At minimum each instrument file requires you to modify the name of the API
function call (e.g. ps6000xxxx vs ps4000xxxx).

This is to force the authors of the instrument files to actually read the
documentation as opposed to assuming similarities between scopes.

You can find pico-python at github.com/colinoflynn/pico-python .
"""

from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import inspect
import time

import numpy as np

from .error_codes import ERROR_CODES as _ERROR_CODES


"""
pico-python is Copyright (c) 2013-2014 By:
Colin O'Flynn <coflynn@newae.com>
Mark Harfouche <mark.harfouche@gmail.com>
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Inspired by Patrick Carle's code at
http://www.picotech.com/support/topic11239.html
which was adapted from
http://www.picotech.com/support/topic4926.html
"""


class _PicoscopeBase(object):
    """
    This class defines a general interface for Picoscope oscilloscopes.

    This  class should not be called directly since it relies on lower level
    functions to communicate with the actual devices.
    """

    #########################################################
    # You must reimplement this in device specific classes
    #########################################################

    # Do not include .dll or .so, these will be appended automatically
    LIBNAME = "ps6000"

    MAX_VALUE = 32764
    MIN_VALUE = -32764

    EXT_MAX_VALUE = 32767
    EXT_MIN_VALUE = -32767
    EXT_RANGE_VOLTS = 20

    CHANNEL_RANGE = [{"rangeV": 20E-3,  "apivalue": 1, "rangeStr": "20 mV"},
                     {"rangeV": 50E-3,  "apivalue": 2, "rangeStr": "50 mV"},
                     {"rangeV": 100E-3, "apivalue": 3, "rangeStr": "100 mV"},
                     {"rangeV": 200E-3, "apivalue": 4, "rangeStr": "200 mV"},
                     {"rangeV": 500E-3, "apivalue": 5, "rangeStr": "500 mV"},
                     {"rangeV": 1.0,    "apivalue": 6, "rangeStr": "1 V"},
                     {"rangeV": 2.0,    "apivalue": 7, "rangeStr": "2 V"},
                     {"rangeV": 5.0,    "apivalue": 8, "rangeStr": "5 V"},
                     ]

    NUM_CHANNELS = 2
    CHANNELS = {"A": 0, "B": 1}

    CHANNEL_COUPLINGS = {"DC50": 2, "DC": 1, "AC": 0}

    BW_LIMITS = {"Full": 0, "20MHZ": 1}

    ############################################################
    # End of things you must reimplement (I think).
    ############################################################

    # If we don't get this CaseInsentiveDict working, I would prefer to stick
    # with their spelling of archaic C all caps for this. I know it is silly,
    # but it removes confusion for certain things like
    # DC_VOLTAGE = DCVoltage or DcVoltage or DC_Voltage
    # or even better
    # SOFT_TRIG = SoftwareTrigger vs SoftTrig

    ERROR_CODES = _ERROR_CODES

    # For some reason this isn't working with me :S
    THRESHOLD_TYPE = {"Above": 0,
                      "Below": 1,
                      "Rising": 2,
                      "Falling": 3,
                      "RiseOrFall": 4}

    # getUnitInfo parameter types
    UNIT_INFO_TYPES = {"DriverVersion": 0x0,
                       "USBVersion": 0x1,
                       "HardwareVersion": 0x2,
                       "VariantInfo": 0x3,
                       "BatchAndSerial": 0x4,
                       "CalDate": 0x5,
                       "KernelVersion": 0x6,
                       "DigitalHardwareVersion": 0x7,
                       "AnalogueHardwareVersion": 0x8,
                       "PicoFirmwareVersion1": 0x9,
                       "PicoFirmwareVersion2": 0xA}

    def __init__(self, serialNumber=None, connect=True):
        """
        Create a scope object, and by default also connect to it.

        Parameters
        ----------
        serialNumber
            Set the serial number in order to connect to a specific scope.
            Leave it empty for connecting to the next free scope.
        connect : bool, default: True
            Connect to the scope
        """
        # TODO: Make A class for each channel
        # that way the settings will make more sense

        # These do not correspond to API values, but rather to
        # the "true" voltage as seen at the oscilloscope probe
        self.CHRange = [5.0] * self.NUM_CHANNELS
        self.CHOffset = [0.0] * self.NUM_CHANNELS
        self.CHCoupling = [1] * self.NUM_CHANNELS
        self.ProbeAttenuation = [1.0] * self.NUM_CHANNELS

        self.handle = None

        if connect is True:
            self.open(serialNumber)

    def getUnitInfo(self, info):
        """Return a string containing information about `info`."""
        if not isinstance(info, int):
            info = self.UNIT_INFO_TYPES[info]
        return self._lowLevelGetUnitInfo(info)

    def getMaxValue(self):
        """Return the maximum ADC value, used for scaling."""
        # TODO: make this more consistent accross versions
        # This was a "fix" when we started supported PS5000a
        return self.MAX_VALUE

    def getMinValue(self):
        """Return the minimum ADC value, used for scaling."""
        return self.MIN_VALUE

    def getAllUnitInfo(self):
        """Return a string containing all information of the device."""
        s = ""
        for key in sorted(self.UNIT_INFO_TYPES.keys(),
                          key=self.UNIT_INFO_TYPES.get):
            s += key.ljust(30) + ": " + self.getUnitInfo(key) + "\n"

        s = s[:-1]
        return s

    def setChannel(self, channel='A', coupling="AC", VRange=2.0,
                   VOffset=0.0, enabled=True, BWLimited=0,
                   probeAttenuation=1.0):
        """
        Set up a specific scope channel with the smallest range possible.

        Parameters
        -----------
        channel
            The channel to set up.
        coupling
            'AC' or 'DC'. Note that if you set a channel to use AC coupling,
            you may need to make a "dummy call" to runBlock, or the first batch
            of data returned via getData* may be inaccurate.
            See https://www.picotech.com/support/topic35401.html
        VRange
            Measurement range in V.
        VOffset
             An offset, in volts, to be added to the input signal before it
             reaches the input amplifier and digitizer.
        enabled
            Enable or disable the channel.
        BWLimited
            Limit the bandwidth.
        probeAttenuation
            See below.

        Return
        ------
        Actual range of the scope as double.


        probeAttenuation
        ----------------

        If using a probe (or a sense resitor), the `probeAttenuation` value is
        used to find the approriate channel range on the scope to use.

        e.g. to use a 10x attenuation probe, you must supply the following
        parameters ps.setChannel('A', "DC", 20.0, 5.0, True, False, 10.0)

        The scope will then be set to use the +- 2V mode at the scope allowing
        you to measure your signal from -25V to +15V.
        After this point, you can set everything in terms of units as seen at
        the tip of the probe. For example, you can set a trigger of 15V and it
        will trigger at the correct value.

        When using a sense resistor, lets say R = 1.3 ohm, you obtain the
        relation:
        V = IR, meaning that your probe as an attenuation of R compared to the
        current you are trying to measure.

        You should supply probeAttenuation = 1.3
        The rest of your units should be specified in amps.

        Unfortunately, you still have to supply a `VRange` that is very close
        to the allowed values. This will change in furture version where we
        will find the next largest range to accomodate the desired range.

        If you want to use units of mA, supply a probe attenuation of 1.3E3.
        Note, the authors recommend sticking to SI units because it makes it
        easier to guess what units each parameter is in.
        """
        if enabled:
            enabled = 1
        else:
            enabled = 0

        if not isinstance(channel, int):
            chNum = self.CHANNELS[channel]
        else:
            chNum = channel

        if not isinstance(coupling, int):
            coupling = self.CHANNEL_COUPLINGS[coupling]

        # finds the next largest range
        VRangeAPI = None
        for item in self.CHANNEL_RANGE:
            if item["rangeV"] - VRange / probeAttenuation > -1E-4:
                if VRangeAPI is None:
                    VRangeAPI = item
                    # break
                # Don't know if this is necessary assuming that it will iterate
                # in order
                elif VRangeAPI["rangeV"] > item["rangeV"]:
                    VRangeAPI = item

        if VRangeAPI is None:
            raise ValueError(
                "Desired range %f is too large. Maximum range is %f." %
                (VRange, self.CHANNEL_RANGE[-1]["rangeV"] * probeAttenuation))

        # store the actually chosen range of the scope
        VRange = VRangeAPI["rangeV"] * probeAttenuation

        if not isinstance(BWLimited, int):
            BWLimited = self.BW_LIMITS[BWLimited]

        if BWLimited == 3:
            BWLimited = 3  # 1MHz Bandwidth Limiter for PicoScope 4444
        elif BWLimited == 2:
            BWLimited = 2  # Bandwidth Limiter for PicoScope 6404,
            # 100kHz Bandwidth Limiter for PicoScope 4444
        elif BWLimited == 1:
            BWLimited = 1  # Bandwidth Limiter for PicoScope 6402/6403,
            # 20kHz Bandwidth Limiter for PicoScope 4444
        else:
            BWLimited = 0

        self._lowLevelSetChannel(chNum, enabled, coupling,
                                 VRangeAPI["apivalue"],
                                 VOffset / probeAttenuation, BWLimited)

        # if all was successful, save the parameters
        self.CHRange[chNum] = VRange
        self.CHOffset[chNum] = VOffset
        self.CHCoupling[chNum] = coupling
        self.ProbeAttenuation[chNum] = probeAttenuation

        return VRange

    def runBlock(self, pretrig=0.0, segmentIndex=0, callback=None):
        """Run a single block.

        Must have already called setSampling for proper setup.

        Parameters
        ----------
        pretrig
            Fraction of samples before the trigger.
        segmentIndex
            Index of scope memory segment to save data to.
        callback
            Function to call once the data acquisition finishes.
        """
        # getting max samples is riddiculous.
        # 1GS buffer means it will take so long
        nSamples = min(self.noSamples, self.maxSamples)

        # to return the same No. Samples ( if pretrig != 0.0 ) I'm wrong ?
        nSamples_pretrig = int(round(nSamples * pretrig))
        self._lowLevelRunBlock(nSamples_pretrig,
                               nSamples - nSamples_pretrig,
                               self.timebase, self.oversample, segmentIndex,
                               callback, None)

    def isReady(self):
        """Return whether scope is ready to transfer data."""
        return self._lowLevelIsReady()

    def waitReady(self, spin_delay=0.01):
        """Block until the scope is ready, retesting every `spin_delay` s."""
        while not self.isReady():
            time.sleep(spin_delay)

    def setSamplingInterval(self, sampleInterval, duration, oversample=0,
                            segmentIndex=0):
        """
        Set the timebase according to a desired sampling interval.

        Parameters
        ----------
        sampleInterval
            Desired interval in seconds between two samples.
        duration
            Desired duration of measurement.
        oversample
            Average over several measurements in the sample interval.
            Not many picoscopes are capable of oversampling.
        segmentIndex
            Index of the memory segment to store data into.

        Return
        ------
        actualSampleInterval : float
            The sample interval in seconds according to the timebase.
        noSamples : int
            Number of samples in the measurement duration.
        maxSamples : int
            Maximum number of samples possible in the memory segment.
        """
        self.oversample = oversample
        self.timebase = self.getTimeBaseNum(sampleInterval)

        timebase_dt = self.getTimestepFromTimebase(self.timebase)

        noSamples = int(round(duration / timebase_dt))

        (self.sampleInterval, self.maxSamples) = self._lowLevelGetTimebase(
            self.timebase, noSamples, oversample, segmentIndex)

        self.noSamples = noSamples
        self.sampleRate = 1.0 / self.sampleInterval
        return (self.sampleInterval, self.noSamples, self.maxSamples)

    def setSamplingFrequency(self, sampleFrequency, noSamples, oversample=0,
                             segmentIndex=0):
        """
        Set the timebase according to a desired sampling frequency.

        Parameters
        ----------
        sampleFrequency
            Desired sample frequency in samples per second.
        noSamples
            Desired number of samples.
        oversample
            Average over several measurements in the sample interval.
            Not many picoscopes are capable of oversampling
        segmentIndex
            Index of the memory segment to store data into.

        Return
        ------
        sampleRate : float
            The sample frequency in samples per second.
        maxSamples : int
            Maximum number of samples possible in the memory segment.
        """
        # TODO: make me more like the functions above
        #       at least in terms of what I return
        sampleInterval = 1.0 / sampleFrequency
        duration = noSamples * sampleInterval
        self.setSamplingInterval(sampleInterval, duration, oversample,
                                 segmentIndex)
        return (self.sampleRate, self.maxSamples)

    def setNoOfCaptures(self, noCaptures):
        """
        Set the number of captures for one run of rapid block mode.

        If you do not call this function before a run, the driver will capture
        one waveform.
        """
        self._lowLevelSetNoOfCaptures(noCaptures)

    def memorySegments(self, noSegments):
        """
        Divide the scope memory into `noSegments` segments.

        The scope fills the available memory segment for a single capture.
        More segments allow several consecutive captures.
        The memory is distributed amongst the active channels: the number of
        samples available to each channel is the maximum number of samples
        divided by 2 (for 2 channels) or 4 (for 3 or 4 channels) or
        8 (for 5 to 8 channels).

        Returns
            Number of samples in the segment.
        """
        maxSamples = self._lowLevelMemorySegments(noSegments)
        self.maxSamples = maxSamples
        self.noSegments = noSegments
        return self.maxSamples

    def getMaxMemorySegments(self):
        """
        Return the maximum number of memory segments allowed by the device.
        """
        segments = self._lowLevelGetMaxSegments()
        return segments

    def setExtTriggerRange(self, VRange=0.5):
        """
        Set the range for the EXT trigger channel.

        This is only implemented for PS4000 series devices where
        the only acceptable values for VRange are 0.5 or 5.0.
        """
        VRangeAPI = None
        for item in self.CHANNEL_RANGE:
            if np.isclose(item["rangeV"], VRange):
                VRangeAPI = item
                break

        if VRangeAPI is None:
            raise ValueError('Provided VRange is not valid')

        self._lowLevelSetExtTriggerRange(VRangeAPI["apivalue"])

    def setSimpleTrigger(self, trigSrc, threshold_V=0, direction="Rising",
                         delay=0, timeout_ms=100, enabled=True):
        """
        Set up a simple trigger.

        Parameters
        ----------
        trigSrc
            Either a channel number corresponding to the low level
            specifications of the scope or a string such as 'A' or 'AUX'.
        threshold_V
            Threshold at which to trigger in V.
        direction
            Can be a text string such as "Rising" or "Falling",
            or the value of the dict from self.THRESHOLD_TYPE[] corresponding
            to your trigger type.
        delay
            Number of clock cycles to wait from trigger conditions met
            until we actually trigger capture.
        timeout_ms
            Time to wait in ms from calling runBlock() or similar
            (e.g. when trigger arms) for the trigger to occur. If no trigger
            occurs it gives up & auto-triggers.
        enabled
            Enable or disable the trigger.

        Notes
        -----
        - Support for offset is currently untested.
        - The AUX port (or EXT) only has a range of +- 1V (at least in PS6000).
        """
        if not isinstance(trigSrc, int):
            trigSrc = self.CHANNELS[trigSrc]

        if not isinstance(direction, int):
            direction = self.THRESHOLD_TYPE[direction]

        if trigSrc >= self.NUM_CHANNELS:
            threshold_adc = int((threshold_V / self.EXT_RANGE_VOLTS) *
                                self.EXT_MAX_VALUE)

            # The external port is typically used as a clock. So I don't think
            # we should raise errors
            threshold_adc = min(threshold_adc, self.EXT_MAX_VALUE)
            threshold_adc = max(threshold_adc, self.EXT_MIN_VALUE)
        else:
            a2v = self.CHRange[trigSrc] / self.getMaxValue()
            threshold_adc = int((threshold_V + self.CHOffset[trigSrc]) / a2v)

            if (threshold_adc > self.getMaxValue() or
               threshold_adc < self.getMinValue()):
                raise IOError(
                    "Trigger Level of %fV outside allowed range (%f, %f)" % (
                        threshold_V,
                        -self.CHRange[trigSrc] - self.CHOffset[trigSrc],
                        +self.CHRange[trigSrc] - self.CHOffset[trigSrc]))

        enabled = int(bool(enabled))

        self._lowLevelSetSimpleTrigger(enabled, trigSrc, threshold_adc,
                                       direction, delay, timeout_ms)

    def getTriggerTimeOffset(self, segmentIndex=0):
        """
        Get the trigger time offset in s for a waveform.

        Parameter
        ---------
        segmentIndex
            Index of the memory segment in question.
        """
        return self._lowLevelGetTriggerTimeOffset(segmentIndex)

    def flashLed(self, times=5, start=False, stop=False):
        """
        Flash the front panel LEDs.

        Parameters
        ----------
        times : int
            Number of times to flash the LED
        start : bool
            Start the LED to flash indefinitely. Equals times < 0.
        stop : bool
            Stop the LED to flash. Equals times == 0

        Note that calls to the RunStreaming or RunBlock will stop any flashing.
        """
        if start:
            times = -1
        if stop:
            times = 0

        self._lowLevelFlashLed(times)

    def getScaleAndOffset(self, channel):
        """
        Return `channel`s scale and offset used to convert the raw waveform.

        To use: first multiply by the scale, then subtract the offset.

        Returns a dictionary with keys 'scale' and 'offset'.
        """
        if not isinstance(channel, int):
            channel = self.CHANNELS[channel]
        return {'scale': self.CHRange[channel] / float(self.getMaxValue()),
                'offset': self.CHOffset[channel]}

    def rawToV(self, channel, dataRaw, dataV=None, dtype=np.float64):
        """
        Convert the raw data to voltage units.

        Parameters
        ----------
        channel
            Scope channel number or name.
        dataRaw
            Raw data.
        dataV
            Numpy array to fill with the data. If None, create an empty one.
        dtype
            Datatype for the numpy array to create.

        Return
        ------
        Numpy array with the measurement in V.
        """
        if not isinstance(channel, int):
            channel = self.CHANNELS[channel]

        if dataV is None:
            dataV = np.empty(dataRaw.shape, dtype=dtype)

        a2v = self.CHRange[channel] / dtype(self.getMaxValue())
        np.multiply(dataRaw, a2v, dataV)
        np.subtract(dataV, self.CHOffset[channel], dataV)

        return dataV

    def getDataV(self, channel, numSamples=0, startIndex=0, downSampleRatio=1,
                 downSampleMode=0, segmentIndex=0, returnOverflow=False,
                 exceptOverflow=False, dataV=None, dataRaw=None,
                 dtype=np.float64):
        """
        Return the data of a single channel as an array of voltage values.

        Paramters
        ---------
        channel
            Scope channel number or name.
        numSamples
            Number of samples to return. If 0, take the calculated number from
            set sampling interval/frequency.
        startIndex
            Index of the sample to start transfer.
        downSampleRatio
            Downsampling factor that will be applied to the raw data. Multiple
            downsampling modes can be bitwise-ORed together, but the
            downSampleRatio must be the same for all modes.
        downSampleMode
            Method of downsampling.
        segmentIndex
            Index of the memory segment to get data from.
        returnOverflow
            Return whether the measurement exceeded the range,
            additionally to the data.
        exceptOverflow
            Raise an IOError exception on overflow.
        dataV
            Numpy array to fill with the data. If None, create an empty one.
        dtype
            Datatype for the numpy array to create.
        dataRaw
            Not used

        Return
        -----
        dataV : numpy array
            Numpy array with the values in Volts.
        overflow : bool, only if returnOverflow is True
            Whether the measured value exceeded the measurement range.
        """
        (dataRaw, numSamplesReturned, overflow) = self.getDataRaw(
            channel, numSamples, startIndex, downSampleRatio, downSampleMode,
            segmentIndex, dataRaw)

        if dataV is None:
            dataV = self.rawToV(channel, dataRaw, dtype=dtype)
            dataV = dataV[:numSamplesReturned]
        else:
            self.rawToV(channel, dataRaw, dataV, dtype=dtype)
            dataV[numSamplesReturned:] = np.nan

        if returnOverflow:
            return (dataV, overflow)
        else:
            if overflow and exceptOverflow:
                raise IOError("Overflow detected in data")
            return dataV

    def getDataRaw(self, channel='A', numSamples=0, startIndex=0,
                   downSampleRatio=1, downSampleMode=0, segmentIndex=0,
                   data=None):
        """
        Return the data of a single channel.

        Paramters
        ---------
        channel
            Scope channel number or name.
        numSamples
            Number of samples to return. If 0, take the calculated number from
            set sampling interval/frequency.
        startIndex
            Index of the sample to start transfer.
        downSampleRatio
            Downsampling factor that will be applied to the raw data. Multiple
            downsampling modes can be bitwise-ORed together, but the
            downSampleRatio must be the same for all modes.
        downSampleMode
            Method of downsampling.
        segmentIndex
            Index of the memory segment to get data from.
        data
            Numpy array to fill with the data. If None, create an empty one.

        Return
        ------
        data : numpy array
            Array with the digitalized measurement values.
        NumSamplesReturned : int
            Number of samples returned.
        overflow : bool
            Whether the measured value exceeded the measurement range.
        """
        if not isinstance(channel, int):
            channel = self.CHANNELS[channel]

        if numSamples == 0:
            # maxSamples is probably huge, 1Gig Sample can be HUGE....
            numSamples = min(self.maxSamples, self.noSamples)

        if data is None:
            data = np.empty(numSamples, dtype=np.int16)
        else:

            if data.dtype != np.int16:
                raise TypeError('Provided array must be int16')
            if data.size < numSamples:
                raise ValueError(
                    'Provided array must be at least as big as numSamples.')
            # see numpy.ndarray.flags
            if data.flags['CARRAY'] is False:
                raise TypeError('Provided array must be c_contiguous,' +
                                ' aligned and writeable.')

        self._lowLevelSetDataBuffer(channel, data, downSampleMode,
                                    segmentIndex)

        (numSamplesReturned, overflow) = self._lowLevelGetValues(
            numSamples, startIndex, downSampleRatio, downSampleMode,
            segmentIndex)
        # necessary or else the next call to getValues will try to fill
        # this array unless it is a call trying to read the same channel.
        self._lowLevelClearDataBuffer(channel, segmentIndex)

        # overflow is a bitwise mask
        overflow = bool(overflow & (1 << channel))

        return (data, numSamplesReturned, overflow)

    def getDataRawBulk(self, channel='A', numSamples=0, fromSegment=0,
                       toSegment=None, downSampleRatio=1, downSampleMode=0,
                       data=None):
        """
        Get one or more waveforms collected in rapid block mode.

        Parameters
        ----------
        channel
            Scope channel number or name.
        numSamples
            Number of samples per segment to return.
            If 0, take the calculated number from set sampling interval.
        fromSegment
            Index of the first segment to retrieve data from.
        toSegment
            Index of the last segment to retrieve data from. If None: the last.
        downSampleRatio
            Downsampling factor that will be applied to the raw data. Multiple
            downsampling modes can be bitwise-ORed together, but the
            downSampleRatio must be the same for all modes.
        downSampleMode
            Method of downsampling.
        data
            Numpy array to fill with the data. If None, create an empty one.

        Return
        ------
        data : numpy array
            Array containing the raw data.
        numSamples : int
            Number of samples retrieved per waveform.
        overflow : numpy array
            Array containing whether the range was exceeded.
        """
        if not isinstance(channel, int):
            channel = self.CHANNELS[channel]
        if toSegment is None:
            toSegment = self.noSegments - 1
        if numSamples == 0:
            numSamples = min(self.maxSamples, self.noSamples)

        numSegmentsToCopy = toSegment - fromSegment + 1
        if data is None:
            data = np.ascontiguousarray(
                np.zeros((numSegmentsToCopy, numSamples), dtype=np.int16))

        # set up each row in the data array as a buffer for one of
        # the memory segments in the scope
        for i, segment in enumerate(range(fromSegment, toSegment + 1)):
            self._lowLevelSetDataBufferBulk(channel,
                                            data[i],
                                            segment,
                                            downSampleMode)
        overflow = np.ascontiguousarray(
            np.zeros(numSegmentsToCopy, dtype=np.int16))

        self._lowLevelGetValuesBulk(numSamples, fromSegment, toSegment,
                                    downSampleRatio, downSampleMode, overflow)

        # don't leave the API thinking these can be written to later
        for i, segment in enumerate(range(fromSegment, toSegment + 1)):
            self._lowLevelClearDataBuffer(channel, segment)

        return (data, numSamples, overflow)

    def setSigGenBuiltInSimple(self,
                               offsetVoltage=0, pkToPk=2, waveType="Sine",
                               frequency=1E6, shots=1, triggerType="Rising",
                               triggerSource="None", stopFrequency=None,
                               increment=10.0, dwellTime=1E-3, sweepType="Up",
                               numSweeps=0):
        """
        Generate simple signals using the built-in waveforms.

        Parameters
        ----------
        offsetVoltage
            Voltage offset.
        pkToPk
            Voltage from peak to peak
        waveType
            Type of waveform: 'Sine', 'Square', 'Triangle', 'RampUp',
            'RampDown', and 'DCVoltage'.
            Some hardware also supports these additional waveforms:
            'Sinc', 'Gaussian', 'HalfSine', and 'WhiteNoise'.
        frequency
            Frequency in Hertz to start at.
        shots
            How often to repeat the waveform. If nonzero, `numSweeps`=0.
        triggerType
            Type of trigger edge/threshold. See SIGGEN_TRIGGER_TYPES.
        triggerSource
            Source of the waveform trigger. See SIGGEN_TRIGGER_SOURCES.
        stopFrequency
            Frequency in Hertz to stop at. If None, do not sweep.
        increment
            Frequency change in Hertz, if in sweep mode.
        dwellTime
            Time in seconds between two frequency changes in sweep mode.
        sweepType
            Type of sweeping the frequency: 'Up', 'Down', 'UpDown', 'DownUp'.
        numSweeps
            Number of sweeps. If nonzero, `shots` has to be zero.
        """
        # I put this here, because the python idiom None is very
        # close to the "None" string we expect
        if triggerSource is None:
            triggerSource = "None"

        if not isinstance(waveType, int):
            waveType = self.WAVE_TYPES[waveType]
        if not isinstance(triggerType, int):
            triggerType = self.SIGGEN_TRIGGER_TYPES[triggerType]
        if not isinstance(triggerSource, int):
            triggerSource = self.SIGGEN_TRIGGER_SOURCES[triggerSource]
        if not isinstance(sweepType, int):
            sweepType = self.SWEEP_TYPES[sweepType]

        self._lowLevelSetSigGenBuiltInSimple(
            offsetVoltage, pkToPk, waveType, frequency, shots, triggerType,
            triggerSource, stopFrequency, increment, dwellTime, sweepType,
            numSweeps)

    def setAWGSimple(self, waveform, duration, offsetVoltage=None,
                     pkToPk=None, indexMode="Single", shots=1,
                     triggerType="Rising", triggerSource="ScopeTrig"):
        """
        Set the AWG to output your desired waveform.

        If you require precise control of the timestep increment, you should
        use setSigGenAritrarySimpleDeltaPhase instead

        Check setSigGenAritrarySimpleDeltaPhase for parameter explanation

        Returns:
            The actual duration of the waveform.
        """
        sampling_interval = duration / len(waveform)

        if not isinstance(indexMode, int):
            indexMode = self.AWG_INDEX_MODES[indexMode]

        if indexMode == self.AWG_INDEX_MODES["Single"]:
            pass
        elif indexMode == self.AWG_INDEX_MODES["Dual"]:
            sampling_interval /= 2
        elif indexMode == self.AWG_INDEX_MODES["Quad"]:
            sampling_interval /= 4

        deltaPhase = self.getAWGDeltaPhase(sampling_interval)

        actual_duration = self.setAWGSimpleDeltaPhase(
            waveform, deltaPhase, offsetVoltage, pkToPk, indexMode, shots,
            triggerType, triggerSource)

        return (actual_duration, deltaPhase)

    def setAWGSimpleDeltaPhase(self, waveform, deltaPhase, offsetVoltage=None,
                               pkToPk=None, indexMode="Single", shots=1,
                               triggerType="Rising",
                               triggerSource="ScopeTrig"):
        """
        Specify deltaPhase between each sample and not the total waveform
        duration.

        Returns the actual time duration of the waveform

        If pkToPk and offset Voltage are both set to None, then their values
        are computed as

        pkToPk = np.max(waveform) - np.min(waveform)
        offset = (np.max(waveform) + np.min(waveform)) / 2

        This should in theory minimize the quantization error in the ADC.

        else, the waveform should be a numpy int16 type array with the
        containing waveform

        For the Quad mode, if offset voltage is not provided, then waveform[0]
        is assumed to be the offset. In quad mode, the offset voltage is the
        point of symmetry

        This is function provides a little more control than
        setAWGSimple in the sense that you are able to specify deltaPhase
        directly. It should only be used when deltaPhase becomes very large.

        Warning. Ideally, you would want this to be a power of 2 that way each
        sample is given out at exactly the same difference in time otherwise,
        if you give it something closer to .75 you would obtain

         T  | phase accumulator value | sample
         0  |      0                  |      0
         5  |      0.75               |      0
        10  |      1.50               |      1
        15  |      2.25               |      2
        20  |      3.00               |      3
        25  |      3.75               |      3

        notice how sample 0 and 3 were played twice  while others were only
        played once.
        This is why this low level function is exposed to the user so that he
        can control these edge cases

        I would suggest using something like this: if you care about obtaining
        evenly spaced samples at the expense of the precise duration of the
        your waveform
        To find the next highest power of 2
            always a smaller sampling interval than the one you asked for
        math.pow(2, math.ceil(math.log(deltaPhase, 2)))

        To find the next smaller power of 2
            always a larger sampling interval than the one you asked for
        math.pow(2, math.floor(math.log(deltaPhase, 2)))

        To find the nearest power of 2
        math.pow(2, int(math.log(deltaPhase, 2), + 0.5))
        """
        """
        This part of the code is written for the PS6403
        (PS6403B if that matters)
        I don't really know a good way to differentiate between PS6403 versions

        It essentially does some autoscaling for the waveform so that it can be
        sent to the Picoscope to allow for maximum resolution from the DDS.

        I haven't tested if you can actually obtain more resolution than simply
        setting the DDS to output from -2 to +2

        I assume they have some type of adjustable gain and offset on their DDS
        allowing them to claim that they can get extremely high resolution.
        """
        if not isinstance(indexMode, int):
            indexMode = self.AWG_INDEX_MODES[indexMode]
        if not isinstance(triggerType, int):
            triggerType = self.SIGGEN_TRIGGER_TYPES[triggerType]
        if not isinstance(triggerSource, int):
            triggerSource = self.SIGGEN_TRIGGER_SOURCES[triggerSource]

        if waveform.dtype == np.int16:
            if offsetVoltage is None:
                offsetVoltage = 0.0
            if pkToPk is None:
                pkToPk = 2.0
                # TODO: make this a per scope function assuming 2.0 V AWG
        else:
            if indexMode == self.AWG_INDEX_MODES["Quad"]:
                # Optimize for the Quad mode.
                """
                Quad mode. The generator outputs the contents of the buffer,
                then on its second pass through the buffer outputs the same
                data in reverse order. On the third and fourth passes
                it does the same but with a negative version of the data. This
                allows you to specify only the first quarter of a waveform with
                fourfold symmetry, such as a sine wave, and let the generator
                fill in the other three quarters.
                """
                if offsetVoltage is None:
                    offsetVoltage = waveform[0]
            else:
                # Nothing to do for the dual mode or the single mode
                if offsetVoltage is None:
                    offsetVoltage = (np.max(waveform) + np.min(waveform)) / 2

            # make a copy of the original data as to not clobber up the array
            waveform = waveform - offsetVoltage
            if pkToPk is None:
                pkToPk = np.max(np.absolute(waveform)) * 2

            # waveform should now be baised around 0
            # with
            #     max(waveform) = +pkToPk/2
            #     min(waveform) = -pkToPk/2
            waveform /= pkToPk

            # waveform should now be a number between -0.5 and +0.5

            waveform += 0.5
            # and now the waveform is between 0 and 1
            # inclusively???

            # now the waveform is properly quantized
            waveform *= (self.AWGMaxVal - self.AWGMinVal)
            waveform += self.AWGMinVal

            waveform.round(out=waveform)

            # convert to an int16 typqe as requried by the function
            waveform = np.array(waveform, dtype=np.int16)

            # funny floating point rounding errors
            waveform.clip(self.AWGMinVal, self.AWGMaxVal, out=waveform)

        self._lowLevelSetAWGSimpleDeltaPhase(
            waveform, deltaPhase, offsetVoltage, pkToPk, indexMode, shots,
            triggerType, triggerSource)

        timeIncrement = self.getAWGTimeIncrement(deltaPhase)
        waveform_duration = timeIncrement * len(waveform)

        if indexMode == self.AWG_INDEX_MODES["Single"]:
            pass
        elif indexMode == self.AWG_INDEX_MODES["Dual"]:
            waveform_duration *= 2
        elif indexMode == self.AWG_INDEX_MODES["Quad"]:
            waveform_duration *= 4

        return waveform_duration

    def getAWGDeltaPhase(self, timeIncrement):
        """
        Return the deltaPhase integer used by the AWG.

        This is useful when you are trying to generate very fast waveforms when
        you are getting close to the limits of your waveform generator.

        For example, the PS6000's DDS phase accumulator increments by
        deltaPhase every AWGDACInterval.
        The top 2**self.AWGBufferAddressWidth bits indicate which sample is
        being output by the DDS.

        """
        samplingFrequency = 1 / timeIncrement
        deltaPhase = int(samplingFrequency / self.AWGDACFrequency *
                         2 ** (self.AWGPhaseAccumulatorSize -
                               self.AWGBufferAddressWidth))
        return deltaPhase

    def getAWGTimeIncrement(self, deltaPhase):
        """
        Return the time between AWG samples given a certain deltaPhase.

        You should use this function in conjunction with
        getAWGDeltaPhase to obtain the actual timestep of AWG.

        """
        samplingFrequency = deltaPhase * self.AWGDACFrequency / (
            2 ** (self.AWGPhaseAccumulatorSize - self.AWGBufferAddressWidth))
        return 1 / samplingFrequency

    def sigGenSoftwareControl(self, state=True):
        """
        Trigger the AWG when configured with software triggering.
        """
        self._lowLevelSigGenSoftwareControl(state)

    def setResolution(self, resolution):
        """For 6000-series, 5000-series or certain 4000-series scopes ONLY,
        sets the resolution.
        """
        self._lowLevelSetDeviceResolution(self.ADC_RESOLUTIONS[resolution])

    def getResolution(self):
        """Get the currently set resolution."""
        return self._lowLevelGetDeviceResolution()

    def enumerateUnits(self):
        """Return connected units as serial numbers as list of strings."""
        return self._lowLevelEnumerateUnits()

    def ping(self):
        """Ping unit to check that the already opened device is connected."""
        return self._lowLevelPingUnit()

    def open(self, serialNumber=None):
        """Open the scope, using `serialNumber` if given."""
        self._lowLevelOpenUnit(serialNumber)

    def openUnitAsync(self, serialNumber=None):
        """Open the scope asynchronously, using `serialNumber` if given."""
        self._lowLevelOpenUnitAsync(serialNumber)

    def openUnitProgress(self):
        """
        Check the progress of the opening procedure.

        Return
        ------
        progress
            Percentage of the opening procedure.
        completed : bool
            Whether the opening has completed.
        """
        return self._lowLevelOpenUnitProgress()

    def close(self):
        """
        Close the scope.

        You should call this yourself because the Python garbage collector
        might take some time.
        """
        if self.handle is not None:
            self._lowLevelCloseUnit()
            self.handle = None

    def stop(self):
        """Stop scope acquisition."""
        self._lowLevelStop()

    def __del__(self):
        self.close()

    def checkResult(self, errorCode):
        """Check result of function calls, raise exception if not 0."""
        # NOTE: This will break some oscilloscopes that are powered by USB.
        # Some of the newer scopes, can actually be powered by USB and will
        # return a useful value. That should be given back to the user.
        # I guess we can deal with these edge cases in the functions themselves
        if errorCode == 0:
            return

        else:
            ecName = self.errorNumToName(errorCode)
            ecDesc = self.errorNumToDesc(errorCode)
            raise IOError('Error calling %s: %s (%s)' % (
                str(inspect.stack()[1][3]), ecName, ecDesc))

    def errorNumToName(self, errorCode):
        """Return the name of the `errorCode` as a string."""
        for t in self.ERROR_CODES:
            if t[0] == errorCode:
                return t[1]

    def errorNumToDesc(self, errorCode):
        """Return the description of the `errorCode` as a string."""
        for t in self.ERROR_CODES:
            if t[0] == errorCode:
                try:
                    return t[2]
                except IndexError:
                    return ""

    def changePowerSource(self, powerstate):
        """Change the powerstate of the scope.

        Valid only for PS54XXA/B?
        """
        # I should probably make an enumerate table for these two cases,
        # but they are in fact just the
        # error codes. Picoscope should have made it a separate enumerate
        # themselves.
        # I'll just keep this hack for now
        if not isinstance(powerstate, int):
            if powerstate == "PICO_POWER_SUPPLY_CONNECTED":
                powerstate = 0x119
            elif powerstate == "PICO_POWER_SUPPLY_NOT_CONNECTED":
                powerstate = 0x11A
        self._lowLevelChangePowerSource(powerstate)
